from itertools import product

import pytest
import torch

from swyft.networks.standardization import OnlineStandardizingLayer


class TestStandardizationLayer:
    bss = [1, 128]
    shapes = [(1,), (5,), (10, 5, 2, 1), (1, 1, 1)]
    means = [2, 139]
    stds = [1, 80]
    stables = [True, False]

    @pytest.mark.parametrize(
        "bs, shape, mean, std, stable", product(bss, shapes, means, stds, stables)
    )
    def test_online_standardization_layer_update(self, bs, shape, mean, std, stable):
        osl = OnlineStandardizingLayer(shape, stable=stable)
        osl.train()
        old_stats = osl.n.clone(), osl.mean.clone(), osl.var.clone(), osl.std.clone()

        data = torch.randn(bs, *shape) * std + mean
        _ = osl(data)

        new_stats = osl.n.clone(), osl._mean.clone(), osl.var.clone(), osl.std.clone()

        assert old_stats != new_stats

    @pytest.mark.parametrize(
        "bs, shape, mean, std, stable", product(bss, shapes, means, stds, stables)
    )
    def test_online_standardization_layer_mean(self, bs, shape, mean, std, stable):
        torch.manual_seed(0)

        osl = OnlineStandardizingLayer(shape, stable=stable)
        osl.train()

        data = torch.randn(bs, *shape) * std + mean
        _ = osl(data)

        assert torch.allclose(osl.mean, data.mean(0))

    @pytest.mark.parametrize(
        "bs, shape, mean, std, stable", product(bss, shapes, means, stds, stables)
    )
    def test_online_standardization_layer_std(self, bs, shape, mean, std, stable):
        torch.manual_seed(0)

        osl = OnlineStandardizingLayer(shape, stable=stable)
        osl.train()

        data = torch.randn(bs, *shape) * std + mean
        _ = osl(data)

        if torch.isnan(data.std(0)).all():
            replacement_std = torch.sqrt(torch.ones_like(osl.std) * osl.epsilon)
            assert torch.allclose(osl.std, replacement_std)
        else:
            assert torch.allclose(osl.std, data.std(0))

    @pytest.mark.parametrize(
        "bs, shape, mean, std, stable", product(bss, shapes, means, stds, stables)
    )
    def test_online_standardization_layer_std_average(
        self, bs, shape, mean, std, stable
    ):
        torch.manual_seed(0)

        osl = OnlineStandardizingLayer(shape, stable=stable, use_average_std=True)
        osl.train()

        data = torch.randn(bs, *shape) * std + mean
        _ = osl(data)

        if torch.isnan(data.std(0)).all():
            replacement_std = torch.sqrt(torch.ones_like(osl.std) * osl.epsilon)
            assert torch.allclose(osl.std, replacement_std.mean())
        else:
            assert torch.allclose(osl.std, data.std(0).mean())


if __name__ == "__main__":
    pass



from itertools import combinations
from typing import List, Tuple

from pandas.core.frame import DataFrame
from toolz import keyfilter

from swyft.types import (
    Array,
    MarginalIndex,
    MarginalToArray,
    MarginalToDataFrame,
    StrictMarginalIndex,
)
from swyft.utils.array import tensor_to_array
from swyft.utils.misc import depth


def get_d_dim_marginal_indices(n_parameters: int, d: int) -> StrictMarginalIndex:
    return tuple(combinations(range(n_parameters), d))


def get_corner_marginal_indices(
    n_parameters: int,
) -> Tuple[StrictMarginalIndex, StrictMarginalIndex]:
    """produce the marginals for a corner plot

    Args:
        n_parameters

    Returns:
        marginal_indices_1d, marginal_indices_2d
    """
    marginal_indices_1d = get_d_dim_marginal_indices(n_parameters, 1)
    marginal_indices_2d = get_d_dim_marginal_indices(n_parameters, 2)
    return marginal_indices_1d, marginal_indices_2d


def tupleize_marginal_indices(marginal_indices: MarginalIndex) -> StrictMarginalIndex:
    """Reformat input marginal_indices into sorted and hashable standard form: tuples of tuples.

    a lone input tuple will be respected as coming from the same marginal
    lists will assumed to be collections of different marginals
    """
    if isinstance(marginal_indices, int):
        out = [marginal_indices]
    elif isinstance(marginal_indices, tuple):
        d = depth(marginal_indices)
        if d == 0:
            raise ValueError("how did this happen?")
        elif d == 1:
            return (marginal_indices,)
        elif d == 2:
            return marginal_indices
        else:
            raise ValueError("marginals can only have two layers of depth, no more.")
    else:
        out = list(marginal_indices)

    for i in range(len(out)):
        if isinstance(out[i], int):
            out[i] = (out[i],)
        else:
            out[i] = tuple(sorted(set(out[i])))
    out = tuple(sorted(out))
    return out


def get_marginal_dim_by_key(key: Tuple[int]) -> int:
    return len(key)


def get_marginal_dim_by_value(value: Array) -> int:
    return value.shape[-1]


def filter_marginals_by_dim(marginals: MarginalToArray, dim: int) -> MarginalToArray:
    assert all(
        isinstance(k, tuple) for k in marginals.keys()
    ), "This function works on tuples of parameters."
    return keyfilter(lambda x: get_marginal_dim_by_key(x) == dim, marginals)


def get_df_from_marginal(v: Array, marginal_index: Tuple[int] = None) -> DataFrame:
    v = tensor_to_array(v)
    if isinstance(marginal_index, int):
        marginal_index = [marginal_index]
    elif marginal_index is None:
        marginal_index = list(range(v.shape[-1]))
    else:
        marginal_index = list(marginal_index)
    return DataFrame(v, columns=marginal_index)


def get_df_dict_from_marginals(marginals: MarginalToArray) -> MarginalToDataFrame:
    return {key: get_df_from_marginal(marginals[key], key) for key in marginals.keys()}



from __future__ import annotations

import logging
from typing import Callable, Optional, TypeVar

import numpy as np
from sklearn.neighbors import BallTree

from swyft.saveable import StateDictSaveable
from swyft.types import ObsType
from swyft.weightedmarginals import WeightedMarginalSamples

log = logging.getLogger(__name__)


BoundType = TypeVar("BoundType", bound="Bound")


class Bound(StateDictSaveable):
    """A bound region on the hypercube.

    .. note::
        The Bound object provides methods to sample from subregions of the
        hypercube, to evaluate the volume of the constrained region, and to
        evaluate the bound.
    """

    def __init__(self) -> None:
        pass

    @property
    def volume(self) -> float:
        """Volume of the bound region."""
        raise NotImplementedError

    @property
    def n_parameters(self) -> int:
        """Number of dimensions."""
        raise NotImplementedError

    def sample(self, n_samples: int) -> np.ndarray:
        """Sample.

        Args:
            n_samples: Numbe of samples.

        Returns:
            s (n_samples x n_parameters)
        """
        raise NotImplementedError

    def __call__(self, u: np.ndarray):
        """Check whether parameters are within bounds.

        Args:
            u (n_samples x n_parameters): Parameters on hypercube
        """
        raise NotImplementedError

    # TODO can we do away with this thing? I think yes.
    @classmethod
    def from_state_dict(cls, state_dict) -> BoundType:
        """Instantiate Bound object based on state_dict.

        Args:
            state_dict (dict): State dictionary
        """
        tag = state_dict["tag"]
        if tag == "UnitCubeBound":
            return UnitCubeBound.from_state_dict(state_dict)
        elif tag == "RectangleBound":
            return RectangleBound.from_state_dict(state_dict)
        elif tag == "BallsBound":
            return BallsBound.from_state_dict(state_dict)
        elif tag == "CompositBound":
            return CompositBound.from_state_dict(state_dict)
        else:
            raise KeyError

    @staticmethod
    def from_marginal_posterior(
        n_samples: int,
        observation: ObsType,
        marginal_posterior: "swyft.inference.marginalposterior.MarginalPosterior",
        threshold: float = -13.0,
        batch_size: Optional[int] = None,
    ) -> BoundType:
        """see CompositBound.from_marginal_posterior"""
        return CompositBound.from_marginal_posterior(
            n_samples,
            observation,
            marginal_posterior,
            threshold,
            batch_size,
        )


class UnitCubeBound(Bound, StateDictSaveable):
    """The unit hypercube bound."""

    def __init__(self, n_parameters):
        """Initialize unit hypercube bound.

        Args:
            n_parameters (int): Number of parameters.
        """
        self._n_parameters = n_parameters
        self._volume = 1.0

    @property
    def volume(self):
        """The volume of the constrained region."""
        return self._volume

    @property
    def n_parameters(self) -> int:
        return self._n_parameters

    def sample(self, n_samples):
        """Generate samples from the bound region.

        Args:
            n_samples (int): Number of samples
        """
        return np.random.rand(n_samples, self.n_parameters)

    def __call__(self, u):
        """Evaluate bound.

        Args:
            u (array): Input array.

        Returns:
            Ones and zeros
        """
        b = np.where(u <= 1.0, np.where(u >= 0.0, 1.0, 0.0), 0.0)
        return b.prod(axis=-1)

    def state_dict(self):
        return dict(tag="UnitCubeBound", n_parameters=self.n_parameters)

    @classmethod
    def from_state_dict(cls, state_dict):
        return cls(state_dict["n_parameters"])


class RectangleBound(Bound, StateDictSaveable):
    def __init__(self, rec_bounds):
        """Rectangle bound.

        Args:
            rec_bounds (n x 2 np.ndarray): list of (u_min, u_max) values.

        Note: 0 <= u_min < u_max  <= 1.
        """
        self._rec_bounds = rec_bounds

    @property
    def volume(self):
        V = 1.0
        for i in range(self.n_parameters):
            V *= self._rec_bounds[i, 1] - self._rec_bounds[i, 0]
        return V

    @property
    def n_parameters(self):
        return len(self._rec_bounds)

    def sample(self, n_samples):
        u = np.random.rand(n_samples, self.n_parameters)
        for i in range(self.n_parameters):
            u[:, i] *= self._rec_bounds[i, 1] - self._rec_bounds[i, 0]
            u[:, i] += self._rec_bounds[i, 0]
        return u

    def __call__(self, u):
        m = np.ones(len(u))
        for i, v in enumerate(self._rec_bounds):
            m *= np.where(u[:, i] >= v[0], np.where(u[:, i] <= v[1], 1.0, 0.0), 0.0)
        return m > 0.5

    def state_dict(self):
        return dict(tag="RectangleBound", rec_bounds=self._rec_bounds)

    @classmethod
    def from_state_dict(cls, state_dict):
        return cls(state_dict["rec_bounds"])


class BallsBound(Bound, StateDictSaveable):
    def __init__(self, points, scale=1.0):
        """Simple mask based on coverage balls around inducing points.

        Args:
            points (Array): shape (num_points, n_dim)
            scale (float): Scale ball size (default 1.0)
        """
        assert len(points.shape) == 2
        self.X = points
        self._n_parameters = self.X.shape[-1]
        self.bt = BallTree(self.X, leaf_size=2)
        self.epsilon = self._set_epsilon(self.X, self.bt, scale)
        self._volume = self._get_volume(self.X, self.epsilon, self.bt)

    @property
    def volume(self) -> float:
        return self._volume

    @property
    def n_parameters(self) -> int:
        return self._n_parameters

    @staticmethod
    def _set_epsilon(X, bt, scale):
        dims = X.shape[-1]
        k = [4, 5, 6]
        dist, ind = bt.query(X, k=k[dims - 1])  # 4th NN
        epsilon = np.median(dist[:, -1]) * scale * 1.5
        return epsilon

    @staticmethod
    def _get_volume(X, epsilon, bt):
        n_samples = 100
        vol_est = []
        d = X.shape[-1]
        area = {1: 2 * epsilon, 2: np.pi * epsilon ** 2}[d]
        for i in range(n_samples):
            n = np.random.randn(*X.shape)
            norm = (n ** 2).sum(axis=1) ** 0.5
            n = n / norm.reshape(-1, 1)
            r = np.random.rand(len(X)) ** (1 / d) * epsilon
            Y = X + n * r.reshape(-1, 1)
            in_bounds = ((Y >= 0.0) & (Y <= 1.0)).prod(axis=1, dtype="bool")
            Y = Y[in_bounds]
            counts = bt.query_radius(Y, epsilon, count_only=True)
            vol_est.append(area * sum(1.0 / counts))
        vol_est = np.array(vol_est)
        out, err = vol_est.mean(), vol_est.std() / np.sqrt(n_samples)
        rel = err / out
        if rel > 0.01:
            log.debug("WARNING: Rel volume uncertainty is %.4g" % rel)
        return out

    def sample(self, n_samples):
        counter = 0
        samples = []
        d = self.X.shape[-1]
        while counter < n_samples:
            n = np.random.randn(*self.X.shape)
            norm = (n ** 2).sum(axis=1) ** 0.5
            n = n / norm.reshape(-1, 1)
            r = np.random.rand(len(self.X)) ** (1 / d) * self.epsilon
            Y = self.X + n * r.reshape(-1, 1)
            in_bounds = ((Y >= 0.0) & (Y <= 1.0)).prod(axis=1, dtype="bool")
            Y = Y[in_bounds]
            counts = self.bt.query_radius(Y, r=self.epsilon, count_only=True)
            p = 1.0 / counts
            w = np.random.rand(len(p))
            Y = Y[p >= w]
            samples.append(Y)
            counter += len(Y)
        samples = np.vstack(samples)
        ind = np.random.choice(range(len(samples)), size=n_samples, replace=False)
        return samples[ind]

    def __call__(self, u):
        u = u.reshape(len(u), -1)
        dist, ind = self.bt.query(u, k=1)
        return (dist < self.epsilon)[:, 0]

    def state_dict(self):
        return dict(
            tag="BallsBound", points=self.X, epsilon=self.epsilon, volume=self._volume
        )

    @classmethod
    def from_state_dict(cls, state_dict):
        obj = cls.__new__(cls)
        obj.X = state_dict["points"]
        assert len(obj.X.shape) == 2
        obj._n_parameters = obj.X.shape[-1]
        obj.epsilon = state_dict["epsilon"]
        obj._volume = state_dict["volume"]
        obj.bt = BallTree(obj.X, leaf_size=2)
        return obj


class CompositBound(Bound, StateDictSaveable):
    """Composit bound object. Product of multiple bounds."""

    def __init__(self, bounds_map, n_parameters):
        """
        Args:
            bounds_map (dict): Dictionary mapping indices like (0, 3) etc --> bounds
            n_parameters (int): Length of parameter vector.
        """
        self._bounds = bounds_map
        self._n_parameters = n_parameters

    def sample(self, n_samples):
        results = -np.ones((n_samples, self.n_parameters))
        for k, v in self._bounds.items():
            results[:, np.array(k)] = v.sample(n_samples)
        return results

    @property
    def volume(self) -> float:
        volume = 1.0
        for k, v in self._bounds.items():
            volume *= v.volume
        return volume

    @property
    def n_parameters(self) -> int:
        return self._n_parameters

    def __call__(self, u):
        res = []
        for k, v in self._bounds.items():
            r = v(u[:, np.array(k)])
            res.append(r)
        return sum(res) == len(res)

    # - Function: Generate sample from posterior
    #   - Constraints are based on p(u|z)/p(u), and should be (different from what we have in the paper???)
    #   - That means I need weights without prior corrections, can be an option to switch this on or off
    #   - Samples should be samples from the hypercube
    #   - Use sampled points above a threshold for generating Rec bound and BallBounds, directly based on points
    # - Function: Return isolated ratio function & bound object from Posteriors object
    #   - Can be used in sampling

    @classmethod
    def from_weighted_samples(
        cls,
        weighted_samples: WeightedMarginalSamples,
        cdf: Callable,
        n_parameters: int,
        threshold: float,
    ) -> BoundType:
        """create a new bound object from weighted samples and the cdf

        Args:
            weighted_samples: log weighted samples
            cdf: transforms from v to u
            n_parameters: number of total parameters
            threshold: above which log weight do we bound? -13 is standard.

        Returns:
            a bound object based on the above
        """
        bounds = {}
        u = cdf(weighted_samples.v)
        for marginal_index in weighted_samples.marginal_indices:
            logw, _ = weighted_samples.get_logweight_marginal(marginal_index)
            mask = logw - logw.max() > threshold
            u_above_threshold = u[mask][:, marginal_index]
            bounds[marginal_index] = BallsBound(u_above_threshold)
        return cls(bounds, n_parameters)

    @classmethod
    def from_marginal_posterior(
        cls,
        n_samples: int,
        observation: ObsType,
        marginal_posterior: "swyft.inference.marginalposterior.MarginalPosterior",
        threshold: float,
        batch_size: Optional[int] = None,
    ) -> BoundType:
        """create a new bound object from a marginal posterior by sampling to estimate the log_prob contours

        Args:
            n_samples: number of samples to estimate with
            observation: single observation to define the bounds
            marginal_posterior: marginal posterior object
            threshold: above which log weight do we bound? -13 is standard
            batch_size: when evaluating the log_prob, what batch size to use

        Returns:
            a bound object based on the above
        """
        weighted_samples = marginal_posterior.weighted_sample(
            n_samples, observation, batch_size
        )
        return cls.from_weighted_samples(
            weighted_samples=weighted_samples,
            cdf=marginal_posterior.prior.cdf,
            n_parameters=marginal_posterior.prior.n_parameters,
            threshold=threshold,
        )

    def state_dict(self):
        state_dict = dict(
            tag="CompositBound",
            n_parameters=self.n_parameters,
            bounds={k: v.state_dict() for k, v in self._bounds.items()},
        )
        return state_dict

    @classmethod
    def from_state_dict(cls, state_dict):
        bounds = {k: Bound.from_state_dict(v) for k, v in state_dict["bounds"].items()}
        n_parameters = state_dict["n_parameters"]
        return cls(bounds, n_parameters)



import numpy as np
import pytest
from matplotlib.figure import Figure

from swyft.plot.histogram import corner, hist1d
from swyft.utils.marginals import filter_marginals_by_dim, get_corner_marginal_indices
from swyft.weightedmarginals import WeightedMarginalSamples


class TestCorner:
    @pytest.mark.parametrize("n_parameters", [2, 3])
    def test_corner_weighted_samples(self, n_parameters: int) -> None:
        n_samples = 1_000
        marginal_indices_1d, marginal_indices_2d = get_corner_marginal_indices(
            n_parameters
        )
        marginal_indices = list(marginal_indices_1d) + list(marginal_indices_2d)
        v = 0.1 * np.random.rand(n_samples, n_parameters) + 0.45
        weights = {
            marginal_index: np.random.rand(n_samples)
            for marginal_index in marginal_indices
        }

        weighted_samples = WeightedMarginalSamples(weights, v)
        fig, axes = corner(
            marginal_1d=weighted_samples.filter_by_dim(1),
            marginal_2d=weighted_samples.filter_by_dim(2),
        )
        assert isinstance(fig, Figure)
        assert axes.shape == (n_parameters, n_parameters)

    @pytest.mark.parametrize("n_parameters", [2, 3])
    def test_corner_samples(self, n_parameters: int) -> None:
        n_samples = 1_000
        marginal_indices_1d, marginal_indices_2d = get_corner_marginal_indices(
            n_parameters
        )
        marginal_indices = list(marginal_indices_1d) + list(marginal_indices_2d)
        marginal_samples = {
            marginal_index: np.random.rand(n_samples, len(marginal_index))
            for marginal_index in marginal_indices
        }

        fig, axes = corner(
            marginal_1d=filter_marginals_by_dim(marginal_samples, 1),
            marginal_2d=filter_marginals_by_dim(marginal_samples, 2),
        )
        assert isinstance(fig, Figure)
        assert axes.shape == (n_parameters, n_parameters)


class TestHist1d:
    @pytest.mark.parametrize("n_parameters", [2, 3])
    def test_hist1d_weighted_samples(self, n_parameters: int) -> None:
        n_samples = 1_000
        marginal_indices, _ = get_corner_marginal_indices(n_parameters)
        v = 0.1 * np.random.rand(n_samples, n_parameters) + 0.45
        weights = {
            marginal_index: np.random.rand(n_samples)
            for marginal_index in marginal_indices
        }

        weighted_samples = WeightedMarginalSamples(weights, v)
        fig, axes = hist1d(
            marginal_1d=weighted_samples,
        )
        assert isinstance(fig, Figure)
        assert axes.shape == (n_parameters,)

    @pytest.mark.parametrize("n_parameters", [2, 3])
    def test_hist1d_samples(self, n_parameters: int) -> None:
        n_samples = 1_000
        marginal_indices, _ = get_corner_marginal_indices(n_parameters)
        marginal_samples = {
            marginal_index: np.random.rand(n_samples, len(marginal_index))
            for marginal_index in marginal_indices
        }

        fig, axes = hist1d(
            marginal_1d=marginal_samples,
        )
        assert isinstance(fig, Figure)
        assert axes.shape == (n_parameters,)



import numpy as np
import pytest
from matplotlib.axes import Axes

from swyft.plot.violin import violin
from swyft.utils.marginals import get_corner_marginal_indices


class TestViolin:
    @pytest.mark.parametrize("n_parameters", [2, 3])
    def test_violin(self, n_parameters: int) -> None:
        n_samples = 1_000
        marginal_indices_1d, _ = get_corner_marginal_indices(n_parameters)
        marginals = {
            marginal_index: np.random.rand(n_samples)
            for marginal_index in marginal_indices_1d
        }

        ax = violin(marginals)
        assert isinstance(ax, Axes)



import numpy as np
import pylab as plt
from scipy.integrate import simps


def grid_interpolate_samples(x, y, bins=1000, return_norm=False):
    idx = np.argsort(x)
    x, y = x[idx], y[idx]
    x_grid = np.linspace(x[0], x[-1], bins)
    y_grid = np.interp(x_grid, x, y)
    norm = simps(y_grid, x_grid)
    y_grid_normed = y_grid / norm
    if return_norm:
        return x_grid, y_grid_normed, norm
    else:
        return x_grid, y_grid_normed


def get_HDI_thresholds(x, cred_level=[0.68268, 0.95450, 0.99730]):
    x = x.flatten()
    x = np.sort(x)[::-1]  # Sort backwards
    total_mass = x.sum()
    enclosed_mass = np.cumsum(x)
    idx = [np.argmax(enclosed_mass >= total_mass * f) for f in cred_level]
    levels = np.array(x[idx])
    return levels


def plot_posterior(
    samples,
    pois,
    weights_key=None,
    ax=plt,
    grid_interpolate=False,
    bins=100,
    color="k",
    contours=True,
    **kwargs
):
    if isinstance(pois, int):
        pois = (pois,)

    w = None

    # FIXME: Clean up ad hoc code
    if weights_key is None:
        weights_key = tuple(sorted(pois))
    try:
        w = samples["weights"][tuple(weights_key)]
    except KeyError:
        if len(weights_key) == 1:
            for k in samples["weights"].keys():
                if weights_key[0] in k:
                    weights_key = k
                    break
            w = samples["weights"][tuple(weights_key)]
        elif len(weights_key) == 2:
            for k in samples["weights"].keys():
                if set(weights_key).issubset(k):
                    weights_key = k
                    w = samples["weights"][k]
    if w is None:
        return

    if len(pois) == 1:
        x = samples["v"][:, pois[0]]

        if grid_interpolate:
            # Grid interpolate samples
            log_prior = samples["log_priors"][pois[0]]
            w_eff = np.exp(np.log(w) + log_prior)  # p(z|x) = r(x, z) p(z)
            zm, v = grid_interpolate_samples(x, w_eff)
        else:
            v, e = np.histogram(x, weights=w, bins=bins, density=True)
            zm = (e[1:] + e[:-1]) / 2

        levels = sorted(get_HDI_thresholds(v))
        if contours:
            contour1d(zm, v, levels, ax=ax, color=color)
        ax.plot(zm, v, color=color, **kwargs)
        ax.set_xlim([x.min(), x.max()])
        ax.set_ylim([-v.max() * 0.05, v.max() * 1.1])

        # Diagnostics
        mean = sum(w * x) / sum(w)
        mode = zm[v == v.max()][0]
        int2 = zm[v > levels[2]].min(), zm[v > levels[2]].max()
        int1 = zm[v > levels[1]].min(), zm[v > levels[1]].max()
        int0 = zm[v > levels[0]].min(), zm[v > levels[0]].max()
        entropy = -simps(v * np.log(v), zm)
        return dict(
            mean=mean, mode=mode, HDI1=int2, HDI2=int1, HDI3=int0, entropy=entropy
        )
    elif len(pois) == 2:
        # FIXME: use interpolation when grid_interpolate == True
        x = samples["v"][:, pois[0]]
        y = samples["v"][:, pois[1]]
        counts, xbins, ybins, _ = ax.hist2d(x, y, weights=w, bins=bins, cmap="gray_r")
        levels = sorted(get_HDI_thresholds(counts))
        try:
            ax.contour(
                counts.T,
                extent=[xbins.min(), xbins.max(), ybins.min(), ybins.max()],
                levels=levels,
                linestyles=[":", "--", "-"],
                colors=color,
            )
        except ValueError:
            print("WARNING: 2-dim contours not well-defined.")
        ax.set_xlim([x.min(), x.max()])
        ax.set_ylim([y.min(), y.max()])

        xm = (xbins[:-1] + xbins[1:]) / 2
        ym = (ybins[:-1] + ybins[1:]) / 2

        cx = counts.sum(axis=1)
        cy = counts.sum(axis=0)

        mean = (sum(xm * cx) / sum(cx), sum(ym * cy) / sum(cy))

        return dict(mean=mean, mode=None, HDI1=None, HDI2=None, HDI3=None, entropy=None)


def plot_1d(
    samples,
    pois,
    truth=None,
    bins=100,
    figsize=(15, 10),
    color="k",
    labels=None,
    label_args={},
    ncol=None,
    subplots_kwargs={},
    fig=None,
    contours=True,
) -> None:
    """Make beautiful 1-dim posteriors.

    Args:
        samples: Samples from `swyft.Posteriors.sample`
        pois: List of parameters of interest
        truth: Ground truth vector
        bins: Number of bins used for histograms.
        figsize: Size of figure
        color: Color
        labels: Custom labels (default is parameter names)
        label_args: Custom label arguments
        ncol: Number of panel columns
        subplot_kwargs: Subplot kwargs
    """

    grid_interpolate = False
    diags = {}

    if ncol is None:
        ncol = len(pois)
    K = len(pois)
    nrow = (K - 1) // ncol + 1

    if fig is None:
        fig, axes = plt.subplots(nrow, ncol, figsize=figsize, **subplots_kwargs)
    else:
        axes = fig.get_axes()
    lb = 0.125
    tr = 0.9
    whspace = 0.15
    fig.subplots_adjust(
        left=lb, bottom=lb, right=tr, top=tr, wspace=whspace, hspace=whspace
    )

    if labels is None:
        labels = [samples["parameter_names"][pois[i]] for i in range(K)]

    for k in range(K):
        if nrow == 1 and ncol > 1:
            ax = axes[k]
        elif nrow == 1 and ncol == 1:
            ax = axes
        else:
            i, j = k % ncol, k // ncol
            ax = axes[j, i]
        ret = plot_posterior(
            samples,
            pois[k],
            ax=ax,
            grid_interpolate=grid_interpolate,
            color=color,
            bins=bins,
            contours=contours,
        )
        ax.set_xlabel(labels[k], **label_args)
        if truth is not None:
            ax.axvline(truth[pois[k]], ls=":", color="r")
        diags[(pois[k],)] = ret
    return fig, diags


def plot_corner(
    samples,
    pois,
    bins=100,
    truth=None,
    figsize=(10, 10),
    color="k",
    labels=None,
    label_args={},
    contours_1d: bool = True,
    fig=None,
) -> None:
    """Make a beautiful corner plot.

    Args:
        samples: Samples from `swyft.Posteriors.sample`
        pois: List of parameters of interest
        truth: Ground truth vector
        bins: Number of bins used for histograms.
        figsize: Size of figure
        color: Color
        labels: Custom labels (default is parameter names)
        label_args: Custom label arguments
        contours_1d: Plot 1-dim contours
        fig: Figure instance
    """
    K = len(pois)
    if fig is None:
        fig, axes = plt.subplots(K, K, figsize=figsize)
    else:
        axes = np.array(fig.get_axes()).reshape((K, K))
    lb = 0.125
    tr = 0.9
    whspace = 0.1
    fig.subplots_adjust(
        left=lb, bottom=lb, right=tr, top=tr, wspace=whspace, hspace=whspace
    )

    diagnostics = {}

    if labels is None:
        labels = [samples["parameter_names"][pois[i]] for i in range(K)]
    for i in range(K):
        for j in range(K):
            ax = axes[i, j]
            # Switch off upper left triangle
            if i < j:
                ax.set_yticklabels([])
                ax.set_xticklabels([])
                ax.set_xticks([])
                ax.set_yticks([])
                ax.set_frame_on(False)
                continue

            # Formatting labels
            if j > 0 or i == 0:
                ax.set_yticklabels([])
                # ax.set_yticks([])
            if i < K - 1:
                ax.set_xticklabels([])
                # ax.set_xticks([])
            if i == K - 1:
                ax.set_xlabel(labels[j], **label_args)
            if j == 0 and i > 0:
                ax.set_ylabel(labels[i], **label_args)

            # Set limits
            # ax.set_xlim(x_lims[j])
            # if i != j:
            #    ax.set_ylim(y_lims[i])

            # 2-dim plots
            if j < i:
                ret = plot_posterior(
                    samples, [pois[j], pois[i]], ax=ax, color=color, bins=bins
                )
                if truth is not None:
                    ax.axvline(truth[pois[j]], color="r")
                    ax.axhline(truth[pois[i]], color="r")
                diagnostics[(pois[j], pois[i])] = ret
            if j == i:
                ret = plot_posterior(
                    samples,
                    pois[i],
                    ax=ax,
                    color=color,
                    bins=bins,
                    contours=contours_1d,
                )
                if truth is not None:
                    ax.axvline(truth[pois[i]], ls=":", color="r")
                diagnostics[(pois[i],)] = ret
    return fig, diagnostics


def contour1d(z, v, levels, ax=plt, linestyles=None, color=None, **kwargs):
    y0 = -1.0 * v.max()
    y1 = 5.0 * v.max()
    ax.fill_between(z, y0, y1, where=v > levels[0], color=color, alpha=0.1)
    ax.fill_between(z, y0, y1, where=v > levels[1], color=color, alpha=0.1)
    ax.fill_between(z, y0, y1, where=v > levels[2], color=color, alpha=0.1)
    # if not isinstance(colors, list):
    #    colors = [colors]*len(levels)
    # for i, l in enumerate(levels):
    #    zero_crossings = np.where(np.diff(np.sign(v-l*1.001)))[0]
    #    for c in z[zero_crossings]:
    #        ax.axvline(c, ls=linestyles[i], color = colors[i], **kwargs)


if __name__ == "__main__":
    pass



from typing import Optional

import pandas as pd
import seaborn as sns
from matplotlib.axes import Axes

from swyft.types import MarginalToArray
from swyft.utils.marginals import filter_marginals_by_dim


def create_violin_df_from_marginal_dict(
    marginals: MarginalToArray, method: Optional[str] = None
) -> pd.DataFrame:
    """map from a marginal sample dict to the df format for violin plots

    Args:
        marginals: marginal dictionary
        method: name of method used to estimate posterior

    Returns:
        violin dataframe
    """
    marginals_1d = filter_marginals_by_dim(marginals, 1)
    rows = []
    for key, value in marginals_1d.items():
        data = {}
        data["Marginal"] = [key[0]] * len(value)
        data["Parameter"] = value.flatten()
        data["Method"] = [method] * len(value)
        df = pd.DataFrame.from_dict(data)
        rows.append(df)
    return pd.concat(rows, ignore_index=True)


def violin(
    marginals: MarginalToArray,
    axes: Axes = None,
    palette: str = "muted",
) -> Axes:
    """create a seaborn violin plot

    Args:
        marginals: marginals from the estimator, must be samples (NOT weighted samples)
        axes: matplotlib axes
        palette: seaborn palette

    Returns:
        Axes
    """
    ax = sns.violinplot(
        x="Marginal",
        y="Parameter",
        data=create_violin_df_from_marginal_dict(marginals),
        palette=palette,
        split=True,
        scale="width",
        inner="quartile",
        ax=axes,
    )
    return ax



import enum
import logging
import os
import shlex
import subprocess
import tempfile
import traceback
from operator import getitem
from typing import Callable, Mapping, Optional, Tuple, Union

import dask.array as da
import numpy as np
import zarr
from dask.distributed import Client, fire_and_forget

from swyft.prior import Prior, PriorTruncator
from swyft.types import ForwardModelType, ObsShapeType, ParameterNamesType, PathType
from swyft.utils import all_finite

log = logging.getLogger(__name__)


class SimulationStatus(enum.IntEnum):
    PENDING = 0
    RUNNING = 1
    FINISHED = 2
    FAILED = 3


class Simulator:
    """Wrapper class for simulator.

    Args:
        model: Model function.
        parameter_names: List of parameter names, or number of parameters (interpreted
            as 'z0', 'z1', ...).
        sim_shapes: Dict describing model function output shapes.
        sim_dtype: Model output data type.
        fail_on_non_finite: whether return an invalid code if simulation
            returns NaN or infinite, default True

    Examples:

        >>> def model(v):
        >>>     mu = sum(v)  # mu = x + y + z
        >>>     nu = np.array([v[1], 2*v[2]])  # nu = [y, 2*z]
        >>>     return dict(mu=mu, nu=nu)
        >>> simulator = swyft.Simulator(model, ["x", "y", "z"], sim_shapes=dict(mu=(1,), nu=(2,))
    """

    def __init__(
        self,
        model: ForwardModelType,
        parameter_names: Union[ParameterNamesType, int],
        sim_shapes: ObsShapeType,
        sim_dtype: str = "f8",
        fail_on_non_finite: bool = True,
    ) -> None:
        self.model = model
        if isinstance(parameter_names, int):
            parameter_names = ["v%i" % i for i in range(parameter_names)]
        self.parameter_names = parameter_names
        self.sim_shapes = sim_shapes
        self.sim_dtype = sim_dtype
        self.fail_on_non_finite = fail_on_non_finite

    def _run(
        self,
        v: Union[zarr.Array, np.ndarray],
        sims: Mapping[str, Union[zarr.indexing.OIndex, np.ndarray]],
        sim_status: Union[zarr.indexing.OIndex, np.ndarray],
        indices: np.ndarray,
        **kwargs
    ) -> None:
        """Run the simulator on the input parameters.

        Args:
            v: Array-like object with all the input parameters. Should have
                shape (num. samples, num. parameters). Should have ``.shape``,
                ``.ndim``, ``.dtype``, and support numpy-style slicing.
            sims: Dictionary of array-like objects where to store the
                simulation output. All arrays should have the number of samples
                as the size of the first dimension. Arrays should support
                numpy-style setitem orthogonal indexing (e.g.
                ``array[:, [1, 2, 3]] = 0``).
            sim_status: Array-like object where to store the simulation status
                (size should be equal to the number of samples). It should
                support numpy-style setitem orthogonal indexing (e.g.
                ``array[:, [1, 2, 3]] = 0``).
            indices: Indices of the samples that need to be run by the
                simulator.
        """
        for i in indices:
            sim, status = _run_model(v[i], self.model, self.fail_on_non_finite)
            for k in sims.keys():
                sims[k][i] = sim.get(k, np.nan)
            sim_status[i] = status

    @classmethod
    def from_command(
        cls,
        command: str,
        parameter_names: Union[ParameterNamesType, int],
        sim_shapes: ObsShapeType,
        set_input_method: Callable,
        get_output_method: Callable,
        shell: bool = False,
        tmpdir: Optional[PathType] = None,
        sim_dtype: str = "f8",
    ):
        """Setup a simulator from a command line program.

        Args:
            command: Command-line program using shell-like syntax.
            set_input_method: Function to setup the simulator input. It should
                take one input argument (the array with the input parameters),
                and return any input to be passed to the program via stdin. If
                the simulator requires any input files to be present, this
                function should write these to disk.
            get_output_method: Function to retrieve results from the simulator
                output. It should take two input arguments (stdout and stderr
                of the simulator run) and return a dictionary with the
                simulator output shaped as described by the ``sim_shapes``
                argument. If the simulator writes output to disk, this function
                should parse the results from the file(s).
            shell: execute the specified command through the shell. NOTE: the
                following security considerations apply:
                https://docs.python.org/3/library/subprocess.html#security-considerations
            tmpdir: Root temporary directory where to run the simulator.
                Each instance of the simulator will run in a separate
                sub-folder. It must exist.
        """
        if shell:
            log.warning(
                "Your command-line program will run through a shell - check the "
                "following security considerations: "
                "https://docs.python.org/3/library/subprocess.html#security-considerations"
            )
        command_args = shlex.split(command) if not shell else command

        def model(v):
            """Closure to setup an instance of the simulator.

            Args:
                v: input parameter array.
            """
            with tempfile.TemporaryDirectory(dir=tmpdir) as tmpdirname:
                cwd = os.getcwd()
                try:
                    os.chdir(tmpdirname)
                    input = set_input_method(v)
                    res = subprocess.run(
                        command_args,
                        capture_output=True,
                        input=input,
                        text=True,
                        check=True,
                        shell=shell,
                    )
                    output = get_output_method(res.stdout, res.stderr)
                finally:
                    os.chdir(cwd)
            return output

        return cls(
            model=model,
            parameter_names=parameter_names,
            sim_shapes=sim_shapes,
            sim_dtype=sim_dtype,
        )

    @classmethod
    def from_model(
        cls, model: ForwardModelType, prior: Prior, fail_on_non_finite: bool = True
    ):
        """Instantiate a Simulator with the correct sim_shapes.

        Args:
            model: Simulator model.
            prior: Model prior.

        Note:
            The simulator model is run once in order to infer observable shapes from the output.
        """
        v = PriorTruncator(prior, bound=None).sample(1)[0]
        sims = model(v)
        sim_shapes = {k: v.shape for k, v in sims.items()}
        dtype = [v.dtype.str for v in sims.values()][0]
        return cls(
            model=model,
            parameter_names=len(v),
            sim_shapes=sim_shapes,
            sim_dtype=dtype,
            fail_on_non_finite=fail_on_non_finite,
        )


class DaskSimulator(Simulator):
    """Setup and run the simulator engine, powered by dask."""

    client = None

    def _run(
        self,
        v: Union[zarr.Array, np.ndarray],
        sims: Mapping[str, Union[zarr.indexing.OIndex, np.ndarray]],
        sim_status: Union[zarr.indexing.OIndex, np.ndarray],
        indices: np.ndarray,
        collect_in_memory: bool = True,
        batch_size: Optional[int] = None,
    ) -> None:
        """Run the simulator on the input parameters.

        Args:
            collect_in_memory: if True, collect the simulation output in
                memory; if False, instruct Dask workers to save the output to
                the corresponding arrays. The latter option is asynchronous,
                thus this method immediately returns.
            batch_size: simulations will be submitted in batches of the
                specified size
        """
        if self.client is None:
            self.set_dask_cluster()

        # open parameter array as Dask array
        chunks = getattr(v, "chunks", "auto")
        v_dask = da.from_array(v, chunks=chunks)
        idx = da.from_array(indices, chunks=(batch_size or -1,))
        v_dask = v_dask[idx]

        v_dask = v_dask.persist()  # load the parameters in the distributed memory

        # block-wise run the model function on the parameter array
        out = da.map_blocks(
            _run_model_chunk,
            v_dask,
            model=self.model,
            sim_shapes=self.sim_shapes,
            fail_on_non_finite=self.fail_on_non_finite,
            drop_axis=1,
            dtype=object,
            meta=np.array((), dtype=object),
        )

        # split result dictionary and simulation status array
        results = out.map_blocks(
            getitem, 0, meta=np.array((), dtype=object), dtype=object
        )
        status = out.map_blocks(getitem, 1, meta=np.array((), dtype=int), dtype=int)

        # unpack array of dictionaries to dictionary of arrays
        result_dict = {}
        for obs, shape in self.sim_shapes.items():
            result_dict[obs] = results.map_blocks(
                getitem,
                obs,
                new_axis=[i + 1 for i in range(len(shape))],
                chunks=(v_dask.chunks[0], *shape),
                meta=np.array(()),
                dtype=float,
            )

        sources = [result_dict[k] for k in self.sim_shapes.keys()]
        targets = [sims[k] for k in self.sim_shapes.keys()]

        if collect_in_memory:
            # submit computation and collect results
            *sources, status = self.client.compute([*sources, status], sync=True)

            # update simulation results
            for source, target in zip(sources, targets):
                target[indices.tolist()] = source

            # finally, update the simulation status
            sim_status[indices.tolist()] = status

        else:
            sources = da.store(
                sources=sources,
                targets=targets,
                regions=(indices.tolist(),),
                lock=False,
                compute=False,
                return_stored=True,
            )

            # submit computation
            *sources, status = self.client.persist([*sources, status])

            # the following dummy array is generated after results are stored.
            zeros_when_done = [
                source.map_blocks(
                    lambda x: np.zeros(x.shape[0], dtype=int),
                    chunks=(source.chunks[0],),
                    drop_axis=[i for i in range(1, source.ndim)],
                    meta=np.array((), dtype=int),
                    dtype=int,
                )
                for source in sources
            ]
            status = sum([*zeros_when_done, status])
            status = status.store(
                target=sim_status,
                regions=(indices.tolist(),),
                lock=False,
                compute=False,
                return_stored=True,
            )
            # when the simulation results are stored, we can update the status
            status = self.client.persist(status)
            fire_and_forget(status)

    def set_dask_cluster(self, cluster=None) -> None:
        """Connect to Dask cluster.

        Args:
            cluster: Cluster address or Cluster object from dask.distributed
                (default is LocalCluster).
        """
        self.client = Client(cluster)


def _run_model(
    v: np.ndarray, model: Callable, fail_on_non_finite: bool
) -> Tuple[Mapping[str, np.ndarray], int]:
    """Run one instance of the model.

    Args:
        v: Array with the model input parameters.
        model: Simulator model function.
        fail_on_non_finite: Whether return an invalid code if simulation
            returns NaN or infinite, default True.

    Return:
        Dictionary with the output of the model run, simulation status code.
    """
    sim = {}
    sim_status = SimulationStatus.FAILED
    try:
        sim = model(v)
        sim_status = _get_sim_status(sim, fail_on_non_finite)
    except:
        traceback.print_exc()
    return sim, sim_status


def _run_model_chunk(
    v: np.ndarray, model: Callable, sim_shapes: ObsShapeType, fail_on_non_finite: bool
) -> Tuple[Mapping[str, np.ndarray], np.ndarray]:
    """Run the model over a set of input parameters.

    Args:
        v: Array with the input parameters. Should have shape (num. samples,
            num. parameters).
        model: Simulator model function.
        sim_shapes: Map of simulator's output names to shapes.
        fail_on_non_finite: Whether return an invalid code if simulation
            returns NaN or infinite, default True.
    Returns:
        Dictionary with the output of the simulations, array with the simulation status.
    """
    chunk_size = len(v)
    sims = {obs: np.full((chunk_size, *shp), np.nan) for obs, shp in sim_shapes.items()}
    sim_status = np.zeros(len(v), dtype=int)
    for i, z_i in enumerate(v):
        sim, status = _run_model(z_i, model, fail_on_non_finite)
        for key, val in sim.items():
            sims[key][i] = val
        sim_status[i] = status
    return sims, sim_status


def _get_sim_status(sims: Mapping[str, np.ndarray], fail_on_non_finite: bool) -> int:
    """Did the simulation fail?"""

    assert isinstance(sims, dict), "Simulators must return a dictionary."

    if any([v is None for v in sims.values()]):
        return SimulationStatus.FAILED
    elif fail_on_non_finite and not all_finite(sims):
        return SimulationStatus.FAILED
    else:
        return SimulationStatus.FINISHED



from abc import ABC, abstractmethod
from typing import Type, TypeVar

import torch

from swyft.types import PathType

StateDictSaveableType = TypeVar("StateDictSaveableType", bound="StateDictSaveable")


class StateDictSaveable(ABC):
    @abstractmethod
    def state_dict(self) -> dict:
        return NotImplementedError()

    @classmethod
    @abstractmethod
    def from_state_dict(cls) -> StateDictSaveableType:
        return NotImplementedError()

    @classmethod
    def load(
        cls: Type[StateDictSaveableType], filename: PathType
    ) -> StateDictSaveableType:
        sd = torch.load(filename)
        return cls.from_state_dict(sd)

    def save(self, filename: PathType) -> None:
        sd = self.state_dict()
        torch.save(sd, filename)


if __name__ == "__main__":
    pass



from swyft.bounds import Bound
from swyft.inference import MarginalPosterior, MarginalRatioEstimator
from swyft.networks import get_marginal_classifier
from swyft.plot import corner, empirical_z_score_corner, hist1d, violin
from swyft.prior import (
    Prior,
    PriorTruncator,
    get_diagonal_normal_prior,
    get_uniform_prior,
)
from swyft.store import DaskSimulator, Dataset, Simulator, Store


def zen():
    print("  Cursed by the dimensionality of your nuisance space?")
    print("  Wasted by Markov chains that reject your simulations?")
    print("     Exhausted from messing with simplistic models,")
    print("because your inference algorithm cannot handle the truth?")
    print("         Try swyft for some pain relief.")


__all__ = [
    "Bound",
    "corner",
    "DaskSimulator",
    "Dataset",
    "empirical_z_score_corner",
    "get_marginal_classifier",
    "hist1d",
    "MarginalPosterior",
    "MarginalRatioEstimator",
    "OnlineDictStandardizingLayer",
    "OnlineStandardizingLayer",
    "Prior",
    "PriorTruncator",
    "Simulator",
    "Store",
    "get_diagonal_normal_prior",
    "get_uniform_prior",
    "violin",
]



from typing import Callable, Dict, Hashable, Optional, Tuple

import torch
import torch.nn as nn
from toolz.dicttoolz import valmap
from torch.utils.data import DataLoader, Dataset, random_split

import swyft
import swyft.utils
from swyft.inference.loss import double_features, double_params
from swyft.inference.train import get_ntrain_nvalid
from swyft.networks.channelized import LinearWithChannel
from swyft.types import Array, Device, MarginalIndex


def is_marginal_block_possible(marginal_indices: MarginalIndex) -> bool:
    marginal_indices = swyft.utils.tupleize_marginals(marginal_indices)
    return [len(marginal_indices[0]) == len(mi) for mi in marginal_indices]


def get_marginal_block_shape(marginal_indices: MarginalIndex) -> Tuple[int, int]:
    marginal_indices = swyft.utils.tupleize_marginals(marginal_indices)
    assert is_marginal_block_possible(
        marginal_indices
    ), f"Each tuple in {marginal_indices} must have the same length."
    return len(marginal_indices), len(marginal_indices[0])


def get_marginal_block(parameters: Array, marginal_indices: MarginalIndex) -> Array:
    depth = swyft.utils.depth(marginal_indices)
    tuple_marginal_indices = swyft.utils.tupleize_marginals(marginal_indices)
    assert is_marginal_block_possible(
        tuple_marginal_indices
    ), f"Each tuple in {tuple_marginal_indices} must have the same length."

    if depth in [0, 1, 2]:
        return torch.stack(
            [parameters[..., mi] for mi in tuple_marginal_indices], dim=1
        )
    else:
        raise ValueError(
            f"{marginal_indices} must be of the form (a) 2, (b) [2, 3], (c) [2, [1, 3]], or (d) [[0, 1], [1, 2]]."
        )


class ObservationTransform(nn.Module):
    def __init__(
        self, observation_key: Hashable, in_features: int, out_features: int
    ) -> None:
        super().__init__()
        self.observation_key = observation_key
        self.in_features = in_features
        self.out_features = out_features
        self.net = nn.Linear(in_features, out_features)

    def forward(self, observation: Dict[Hashable, torch.Tensor]) -> torch.Tensor:
        x = observation[self.observation_key].flatten(1)  # B, O
        return self.net(x)  # B, F


class ParameterTransform(nn.Module):
    def __init__(self, marginal_indices: MarginalIndex):
        super().__init__()
        self.register_buffer(
            "marginal_indices",
            torch.tensor(swyft.utils.tupleize_marginals(marginal_indices)),
        )

    def forward(self, parameters: torch.Tensor) -> torch.Tensor:
        return get_marginal_block(parameters, self.marginal_indices)  # B, M, P


class Classifier(nn.Module):
    def __init__(self, n_marginals: int, n_combined_features: int) -> None:
        super().__init__()
        self.n_marginals = n_marginals
        self.n_combined_features = n_combined_features
        self.net = LinearWithChannel(self.n_marginals, self.n_combined_features, 1)

    def forward(
        self, features: torch.Tensor, marginal_block: torch.Tensor
    ) -> torch.Tensor:
        fb = features.unsqueeze(1).expand(-1, self.n_marginals, -1)  # B, M, F
        combined = torch.cat([fb, marginal_block], dim=2)  # B, M, F + P
        return self.net(combined).squeeze(-1)  # B, M


class Network(nn.Module):
    def __init__(
        self,
        observation_transform: nn.Module,
        parameter_transform: nn.Module,
        classifier: nn.Module,
    ) -> None:
        super().__init__()
        self.observation_transform = observation_transform
        self.parameter_transform = parameter_transform
        self.classifier = classifier

    def forward(
        self, observation: Dict[Hashable, torch.Tensor], parameters: torch.Tensor
    ) -> torch.Tensor:
        features = self.observation_transform(observation)  # B, F
        marginal_block = self.parameter_transform(parameters)  # B, M, P
        return self.classifier(features, marginal_block)  # B, M


def get_classifier(
    observation_key: Hashable,
    marginal_indices: MarginalIndex,
    n_observation_features: int,
    n_observation_embedding_features: int,
) -> nn.Module:
    nm, nbp = get_marginal_block_shape(marginal_indices)
    n_observation_embedding_features = 50
    return Network(
        ObservationTransform(
            observation_key, n_observation_features, n_observation_embedding_features
        ),
        ParameterTransform(marginal_indices),
        Classifier(nm, n_observation_embedding_features + nbp),
    )


################################
# MRE
################################


class MarginalRatioEstimator:
    def __init__(
        self,
        dataset: Dataset,
        marginal_indices: MarginalIndex,
        network: nn.Module,
        device: Device,
    ) -> None:
        self.dataset = dataset
        self.marginal_indices = marginal_indices
        self.network = network
        self.device = device

        self.optimizer = None
        self.scheduler = None
        self.epoch = None
        self.best_state_dict = None
        self.min_loss = float("-Inf")

    def train(
        self,
        batch_size: int = 50,
        learning_rate: float = 5e-4,
        validation_percentage: float = 0.1,
        optimizer: Callable = torch.optim.Adam,
        scheduler: Optional[Callable] = torch.optim.lr_scheduler.ReduceLROnPlateau,
        scheduler_kwargs: dict = {"factor": 0.1, "patience": 5},
        early_stopping_patience: Optional[int] = None,
        max_epochs: int = 2 ** 31 - 1,
        nworkers: int = 0,
        non_blocking: bool = True,
        pin_memory: bool = True,
    ) -> None:
        if early_stopping_patience is None:
            early_stopping_patience = max_epochs

        if self.optimizer is None:
            self.optimizer = optimizer(self.network.parameters(), lr=learning_rate)

        if scheduler is not None and self.scheduler is None:
            self.scheduler = scheduler(self.optimizer, **scheduler_kwargs)

        n_train, n_valid = get_ntrain_nvalid(validation_percentage, len(self.dataset))
        dataset_train, dataset_valid = random_split(self.dataset, [n_train, n_valid])
        train_loader = DataLoader(
            dataset_train,
            batch_size=batch_size,
            num_workers=nworkers,
            pin_memory=pin_memory,
            drop_last=True,
        )
        valid_loader = DataLoader(
            dataset_valid,
            batch_size=min(batch_size, n_valid),
            num_workers=nworkers,
            pin_memory=pin_memory,
            drop_last=True,
        )

        n_validation_batches = len(valid_loader) if len(valid_loader) != 0 else 1
        validation_losses = []
        self.epoch, fruitless_epoch = 0, 0
        while self.epoch < max_epochs and fruitless_epoch < early_stopping_patience:
            # Training
            self.network.train()
            for observation, u, _ in train_loader:
                self.optimizer.zero_grad()
                observation = swyft.utils.dict_to_device(
                    observation, device=self.device, non_blocking=non_blocking
                )
                u = u.to(self.device)
                loss = self._loss(observation, u).sum(dim=0)
                loss.backward()
                self.optimizer.step()

            self.epoch += 1

            # Evaluation
            self.network.eval()
            loss_sum = 0
            with torch.inference_mode():
                for observation, u, _ in valid_loader:
                    observation = swyft.utils.dict_to_device(
                        observation, device=self.device, non_blocking=non_blocking
                    )
                    u = u.to(self.device)
                    validation_loss = self._loss(observation, u).sum(dim=0)
                    loss_sum += validation_loss
                loss_avg = loss_sum / n_validation_batches
                print(
                    "\rtraining: lr=%.2g, epoch=%i, validation loss=%.4g"
                    % (learning_rate, self.epoch, loss_avg),
                    end="",
                    flush=True,
                )
                validation_losses.append(loss_avg)

                if self.epoch == 0 or self.min_loss > validation_loss:
                    fruitless_epoch = 0
                    self.min_loss = validation_loss
                    self.best_state_dict = self.network.state_dict()
                else:
                    fruitless_epoch += 1

                if self.scheduler is not None:
                    self.scheduler.step(loss_avg)

    def _loss(
        self, observation: Dict[Hashable, torch.Tensor], parameters: torch.Tensor
    ) -> torch.Tensor:
        """Returns marginal-wise losses."""
        n_batch = parameters.size(0)
        assert (
            n_batch % 2 == 0
        ), "Loss function can only handle even-numbered batch sizes."
        assert all(
            [v.size(0) == n_batch for v in observation.values()]
        ), "The observation batch_size must agree with the parameter batch_size."

        # Repeat interleave
        observation_doubled = valmap(double_features, observation)
        parameters_doubled = double_params(parameters)

        lnL = self.network(observation_doubled, parameters_doubled)
        lnL = lnL.view(-1, 4, lnL.shape[-1])

        loss = -torch.nn.functional.logsigmoid(lnL[:, 0])
        loss += -torch.nn.functional.logsigmoid(-lnL[:, 1])
        loss += -torch.nn.functional.logsigmoid(-lnL[:, 2])
        loss += -torch.nn.functional.logsigmoid(lnL[:, 3])
        loss = loss.sum(axis=0) / n_batch

        return loss


####################
# Simple Data
####################


class Data(torch.utils.data.Dataset):
    def __init__(self, x, theta) -> None:
        super().__init__()
        self.x = x
        self.theta = theta

    def __getitem__(self, index):
        return (
            {"x": self.x["x"][index]},  # x
            self.theta[index],  # u
            self.theta[index],  # v
        )

    def __len__(self):
        return len(self.theta)


def main():
    b, o = 10, 7
    f = 10
    p = 5
    marginal_indices = [1, 2, 3]
    key = "x"
    x = {key: torch.rand(b, o)}
    theta = torch.rand(b, p)

    device = "cpu"

    N = 10_000
    dataset = Data({"x": torch.rand(N, o)}, torch.rand(N, p))

    network = get_classifier(key, marginal_indices, o, f)
    val = network(x, theta)

    mre = MarginalRatioEstimator(
        dataset,
        marginal_indices,
        network,
        device,
    )
    mre.train()


if __name__ == "__main__":
    main()



from typing import Optional, Tuple, TypeVar
from warnings import warn

import numpy as np
import torch
from torch.types import Device
from torch.utils.data.dataset import Dataset

from swyft.bounds import Bound
from swyft.inference.marginalratioestimator import MarginalRatioEstimator
from swyft.prior import Prior, PriorTruncator
from swyft.types import Array, MarginalIndex, MarginalToArray, ObsType
from swyft.utils import tupleize_marginal_indices
from swyft.weightedmarginals import WeightedMarginalSamples

MarginalPosteriorType = TypeVar("MarginalPosteriorType", bound="MarginalPosterior")


class MarginalPosterior:
    def __init__(
        self,
        marginal_ratio_estimator: MarginalRatioEstimator,
        prior: Prior,
        bound: Optional[Bound] = None,
    ) -> None:
        """a trained marginal ratio estimator and a prior allows access to the posterior

        Args:
            marginal_ratio_estimator: trained marginal ratio estimator
            prior: associated prior. if swyft.Prior is provided then create an untruncated prior to sample from.
        """
        super().__init__()
        self.marginal_ratio_estimator = marginal_ratio_estimator
        self.prior = prior
        self.bound = bound
        self.prior_truncator = PriorTruncator(self.prior, self.bound)

    @property
    def marginal_indices(self) -> MarginalIndex:
        return tupleize_marginal_indices(self.marginal_ratio_estimator.marginal_indices)

    @property
    def device(self) -> Device:
        return self.marginal_ratio_estimator.device

    def truncate(
        self, n_samples: int, observation: ObsType, threshold: float = -13.0
    ) -> Bound:
        return Bound.from_marginal_posterior(
            n_samples=n_samples,
            observation=observation,
            marginal_posterior=self,
            threshold=threshold,
        )

    def empirical_mass(
        self,
        n_observations: int,
        n_posterior_samples: int,
        dataset: Dataset,
    ) -> Tuple[MarginalToArray, MarginalToArray]:
        """compute the empirical and nominal masses

        Args:
            n_observations: number of observations to estimate with
            n_posterior_samples: number of samples to estimate the mass for a certain observation
            dataset: indexable torch dataset which outputs `observation, u, v = dataset[i]`

        Returns:
            empirical mass dict and nominal mass dict for every marginal
        """
        empirical_mass = {marginal: [] for marginal in self.marginal_indices}
        nominal_mass = {
            marginal: np.linspace(1 / n_observations, 1, n_observations)
            for marginal in self.marginal_indices
        }

        for _ in range(n_observations):
            ind = np.random.randint(n_observations)
            observation_o, _, v_o = dataset[ind]
            logw_o = self.marginal_ratio_estimator.log_ratio(
                observation_o, torch.atleast_2d(v_o)
            )
            logw_s = self.weighted_sample(n_posterior_samples, observation_o)

            w_o = {
                marginal_index: np.exp(logw) for marginal_index, logw in logw_o.items()
            }
            w_s = {
                marginal_index: np.exp(logw)
                for marginal_index, logw in logw_s.weights.items()
            }
            for marginal_index, weight_o in w_o.items():
                is_above_true_param_density = w_s[marginal_index] >= weight_o
                sum_above_true_param_density = w_s[marginal_index][
                    is_above_true_param_density
                ].sum()
                percent_above_true_param_density = (
                    sum_above_true_param_density / w_s[marginal_index].sum()
                )
                empirical_mass[marginal_index].append(percent_above_true_param_density)
        for marginal_index, mass in empirical_mass.items():
            empirical_mass[marginal_index] = np.asarray(sorted(mass))

        return empirical_mass, nominal_mass

    def log_prob(
        self, observation: ObsType, v: Array, batch_size: Optional[int] = None
    ) -> MarginalToArray:
        """estimate the log posterior probability of an observation and parameters v

        Args:
            observation: a single observation
            v: parameters to estimate the log proability
            batch_size: batch_size to estimate the log ratio. None attempts to estimate all log ratios at once

        Returns:
            dictionary with marginal_indices keys and log posterior values.
        """
        log_prior = np.atleast_2d(self.prior_truncator.log_prob(v))
        marginal_log_prior = {
            k: log_prior[..., i] for i, k in enumerate(self.marginal_indices)
        }
        log_ratios = self.marginal_ratio_estimator.log_ratio(
            observation, v, batch_size=batch_size
        )
        return {
            marginal_index: log_ratios[marginal_index]
            + marginal_log_prior[marginal_index]
            for marginal_index in self.marginal_indices
        }

    def weighted_sample(
        self, n_samples: int, observation: ObsType, batch_size: Optional[int] = None
    ) -> WeightedMarginalSamples:
        """sample from the prior and estimate the log ratio weights

        Args:
            n_samples: number of desired samples
            observation: observation in the posterior
            batch_size: batch_size to estimate the log ratio. None attempts to estimate all log ratios at once

        Returns:
            dictionary with keys = marginal_indices U {"v"}
                "v" maps to the parameters drawn from the prior (n_samples, n_parameters).
                each marginal index maps to the log ratio (n_samples, len(marginal_index)).
        """
        v = self.prior_truncator.sample(n_samples)
        logweight = self.marginal_ratio_estimator.log_ratio(
            observation, v, batch_size=batch_size
        )
        return WeightedMarginalSamples(logweight, v)

    def sample(
        self,
        n_samples: int,
        observation: ObsType,
        batch_size: Optional[int] = None,
        excess_factor: int = 10,
        maxiter: int = 1000,
    ) -> MarginalToArray:
        """sample from the posterior using rejection sampling

        Args:
            n_samples: number of samples in each marginal to output
            observation: target observation
            batch_size: batch_size to estimate the log ratio. None attempts to estimate all log ratios at once
            excess_factor: n_samples_to_reject = excess_factor * n_samples
            maxiter: maximum loop attempts to draw n_samples

        Returns:
            Marginal posterior samples. keys are marginal tuples, values are samples.

        Reference:
            Section 23.3.3
            Machine Learning: A Probabilistic Perspective
            Kevin P. Murphy
        """
        # From an initial draw, estimate the log MAP
        weighted_samples = self.weighted_sample(
            excess_factor * n_samples, observation, batch_size
        )

        marginal_log_MAP = {
            marginal_index: weighted_samples.get_logweight(marginal_index).max()
            for marginal_index in self.marginal_indices
        }

        # define the work left to do, the place to put it, and the end result / output.
        remaining_marginal_indices = set(marginal_log_MAP.keys())
        collector = {k: [] for k in remaining_marginal_indices}
        out = {}

        # Do the rejection sampling.
        # When a particular key hits the necessary samples, stop calculating on it to reduce cost.
        # Send that key to out.
        counter = 0
        while counter < maxiter:
            # Calculate chance to keep a sample
            log_prob_to_keep = {
                marginal_index: weighted_samples.get_logweight(marginal_index)
                - marginal_log_MAP[marginal_index]
                for marginal_index in remaining_marginal_indices
            }

            # Draw and determine if samples are kept
            to_keep = {
                marginal_index: np.less_equal(
                    np.log(np.random.rand(*log_prob.shape)), log_prob
                )
                for marginal_index, log_prob in log_prob_to_keep.items()
            }

            # Collect samples for every tuple of parameters, if there are enough, add them to out.
            for marginal_index in remaining_marginal_indices:
                all_parameters_to_keep = weighted_samples.v[to_keep[marginal_index]]
                marginal_to_keep = all_parameters_to_keep[..., marginal_index]
                collector[marginal_index].append(marginal_to_keep)
                concatenated = np.concatenate(collector[marginal_index])[:n_samples]
                if len(concatenated) == n_samples:
                    out[marginal_index] = concatenated

            # Remove the param_tuples which we already have in out, thus to avoid calculating them anymore.
            for marginal_index in out.keys():
                if marginal_index in remaining_marginal_indices:
                    remaining_marginal_indices.remove(marginal_index)

            if len(remaining_marginal_indices) > 0:
                weighted_samples = self.weighted_sample(
                    excess_factor * n_samples, observation, batch_size
                )
            else:
                return out
            counter += 1
        warn(
            f"Max iterations {maxiter} reached there were not enough samples produced in {remaining_marginal_indices}."
        )
        return out


if __name__ == "__main__":
    pass



import numpy as np
import pytest

from swyft.plot.constraint import diagonal_constraint, lower_constraint
from swyft.plot.histogram import corner
from swyft.utils.marginals import get_corner_marginal_indices
from swyft.weightedmarginals import WeightedMarginalSamples


class TestConstraint:
    @pytest.mark.skip
    def test_constraint(self):
        n_samples = 1_000
        n_parameters = 3

        marginal_indices_1d, marginal_indices_2d = get_corner_marginal_indices(
            n_parameters
        )
        marginal_indices = list(marginal_indices_1d) + list(marginal_indices_2d)
        v = 0.1 * np.random.rand(n_samples, n_parameters) + 0.45
        weights = {
            marginal_index: np.random.rand(n_samples)
            for marginal_index in marginal_indices
        }
        df_dict = WeightedMarginalSamples(weights, v).get_df_dict()

        bounds = np.asarray([[0.1, 0.9], [0.2, 0.8], [0.3, 0.9]])
        boundss = [bounds, [1.2, 0.8] * bounds]

        _, axes = corner(df_dict)
        for bounds in boundss:
            diagonal_constraint(axes, bounds)
        for bounds in boundss:
            lower_constraint(axes, bounds)



from typing import Callable, Hashable, Optional, Sequence, Tuple
from warnings import warn

import numpy as np
import torch

import swyft
from swyft.saveable import StateDictSaveable
from swyft.types import Array, ObsType, ParameterNamesType, PathType
from swyft.utils.array import array_to_tensor


class SimpleDataset(torch.utils.data.Dataset):
    """Dataset which merely keeps track of observation and parameter data in one place"""

    def __init__(self, observations: ObsType, us: Array, vs: Array) -> None:
        """
        Args:
            observations: dictionary of batched obserations
            us: array or tensor of unit cube parameters
            vs: array or tensor of natural parameters
        """
        super().__init__()
        b = us.shape[0]
        assert (
            vs.shape[0] == b
        ), "the us and vs arrays do not have the same batch dimension"
        assert all(
            [b == x.shape[0] for x in observations.values()]
        ), "the observation values do not have the same batch dimension as us and vs"
        self.observations = observations
        self.us = us
        self.vs = vs
        self._len = b

    def __getitem__(self, idx) -> Tuple[ObsType, torch.Tensor, torch.Tensor]:
        return (
            {
                key: array_to_tensor(val[idx, ...])
                for key, val in self.observations.items()
            },
            array_to_tensor(self.us[idx, ...]),
            array_to_tensor(self.vs[idx, ...]),
        )

    def __len__(self) -> int:
        return self._len


class Dataset(torch.utils.data.Dataset, StateDictSaveable):
    """Dataset for access to swyft.Store."""

    def __init__(
        self,
        N: int,
        prior: swyft.Prior,
        store: "swyft.store.store.Store",
        bound: Optional[swyft.Bound] = None,
        simhook: Optional[Callable[..., ObsType]] = None,
        simkeys: Optional[Sequence[Hashable]] = None,
    ) -> None:
        """
        Args:
            N: Number of samples.
            prior: Parameter prior.
            store: Store reference.
            simhook: Applied on-the-fly to each sample. simhook(x, v)
            simkeys: List of simulation keys that should be exposed (None means that all store sims are exposed)

        .. note::
            swyft.Dataset is essentially a list of indices that point to
            corresponding entries in the swyft.Store.  It is a daughter class of
            torch.utils.data.Dataset, and can be used directly for training.  Due
            to the statistical nature of the Store, the returned number of samples
            is effectively drawn from a Poisson distribution with mean N.
        """
        super().__init__()
        self._prior_truncator = swyft.PriorTruncator(prior, bound)
        self.indices = store.sample(N, prior, bound=bound)
        self._store = store
        self._simhook = simhook
        self._simkeys = simkeys if simkeys else list(self._store.sims)

        if self.requires_sim:
            warn(
                "The store requires simulation for some of the indices in this Dataset."
            )

    def __len__(self) -> int:
        """Return length of dataset."""
        return len(self.indices)

    @property
    def prior(self) -> swyft.Prior:
        """Return prior of dataset."""
        return self._prior_truncator.prior

    @property
    def bound(self) -> swyft.Bound:
        """Return bound of truncated prior of dataset (swyft.Bound)."""
        return self._prior_truncator.bound

    def simulate(
        self, batch_size: Optional[int] = None, wait_for_results: bool = True
    ) -> None:
        """Trigger simulations for points in the dataset.

        Args:
            batch_size: Number of batched simulations.
            wait_for_results: What for simulations to complete before returning.
        """
        self._store.simulate(
            self.indices, batch_size=batch_size, wait_for_results=wait_for_results
        )

    @property
    def requires_sim(self) -> bool:
        """Check if simulations are required for points in the dataset."""
        return self._store.requires_sim(self.indices)

    @property
    def v(self) -> np.ndarray:
        """Return all parameters as (n_points, n_parameters) array."""
        return np.array([self._store.v[i] for i in self.indices])

    @property
    def parameter_names(self) -> ParameterNamesType:
        """Return parameter names (inherited from store and simulator)."""
        return self._store.parameter_names

    def __getitem__(self, idx) -> Tuple[ObsType, torch.Tensor, torch.Tensor]:
        """Return datastore entry."""
        i = self.indices[idx]
        x = {k: self._store.sims[k][i] for k in self._simkeys}
        v = self._store.v[i]
        if self._simhook is not None:
            x = self._simhook(x, v)
        u = self.prior.cdf(v.reshape(1, -1)).flatten()

        return (
            {key: array_to_tensor(val) for key, val in x.items()},
            array_to_tensor(u),
            array_to_tensor(v),
        )

    def state_dict(self) -> dict:
        return dict(
            indices=self.indices,
            prior_truncator=self._prior_truncator.state_dict(),
            simhook=bool(self._simhook),
            simkeys=self._simkeys,
        )

    @classmethod
    def from_state_dict(cls, state_dict, store, simhook=None):
        obj = cls.__new__(cls)
        obj._prior_truncator = swyft.PriorTruncator.from_state_dict(
            state_dict["prior_truncator"]
        )
        obj.indices = state_dict["indices"]

        obj._store = store
        obj._simhook = simhook
        obj._simkeys = state_dict["simkeys"]
        if simhook is None and state_dict["simhook"] is True:
            warn(
                "A simhook was specified when the dataset was saved, but is missing now."
            )
        elif simhook is not None and state_dict["simhook"] is False:
            warn(
                "A simhook was specified, but no simhook was specified when the Dataset was saved."
            )
        return obj

    def save(self, filename: PathType) -> None:
        """
        .. note::
            The store and the simhook are not saved. They must be loaded independently by the user.
        """
        sd = self.state_dict()
        torch.save(sd, filename)

    @classmethod
    def load(
        cls,
        filename: PathType,
        store: "swyft.Store",
        simhook: Callable[..., ObsType] = None,
    ):
        """Load dataset.

        Args:
            filename
            store: Corresponding datastore.
            simhook: Simulation hook.

        .. warning::
            Make sure that the store is the same that was originally used for
            creating the dataset.
        """
        sd = torch.load(filename)
        return cls.from_state_dict(sd, store, simhook=simhook)



from abc import ABC, abstractmethod
from typing import Dict, Hashable, Tuple

import torch
import torch.nn as nn

import swyft
import swyft.utils
from swyft.networks.channelized import ResidualNetWithChannel
from swyft.networks.standardization import (
    OnlineDictStandardizingLayer,
    OnlineStandardizingLayer,
)
from swyft.types import Array, MarginalIndex, ObsShapeType


class HeadTailClassifier(ABC):
    """Abstract class which ensures that child classifier networks will function with swyft"""

    @abstractmethod
    def head(self, observation: Dict[Hashable, torch.Tensor]) -> torch.Tensor:
        """convert the observation into a tensor of features

        Args:
            observation: observation type

        Returns:
            a tensor of features which can be utilized by tail
        """
        pass

    @abstractmethod
    def tail(self, features: torch.Tensor, parameters: torch.Tensor) -> torch.Tensor:
        """finish the forward pass using features computed by head

        Args:
            features: output of head
            parameters: the parameters normally given to forward pass

        Returns:
            the same output as `forward(observation, parameters)`
        """
        pass


class ObservationTransform(nn.Module):
    def __init__(
        self,
        observation_key: Hashable,
        observation_shapes: ObsShapeType,
        online_z_score: bool,
    ) -> None:
        super().__init__()
        self.observation_key = observation_key
        self.observation_shapes = observation_shapes
        self.flatten = nn.Flatten()
        if online_z_score:
            self.online_z_score = OnlineDictStandardizingLayer(self.observation_shapes)
        else:
            self.online_z_score = nn.Identity()

    def forward(self, observation: Dict[Hashable, torch.Tensor]) -> torch.Tensor:
        z_scored_observation = self.online_z_score(observation)
        return self.flatten(z_scored_observation[self.observation_key])  # B, O

    @property
    def n_features(self) -> int:
        with torch.no_grad():
            fabricated_observation = {
                key: torch.rand(2, *shape)
                for key, shape in self.observation_shapes.items()
            }
            _, n_features = self.forward(fabricated_observation).shape
        return n_features


class ParameterTransform(nn.Module):
    def __init__(
        self, n_parameters: int, marginal_indices: MarginalIndex, online_z_score: bool
    ) -> None:
        super().__init__()
        self.register_buffer(
            "marginal_indices",
            torch.tensor(swyft.utils.tupleize_marginal_indices(marginal_indices)),
        )
        self.n_parameters = torch.Size([n_parameters])
        if online_z_score:
            self.online_z_score = OnlineStandardizingLayer(self.n_parameters)
        else:
            self.online_z_score = nn.Identity()

    def forward(self, parameters: torch.Tensor) -> torch.Tensor:
        parameters = self.online_z_score(parameters)
        return self.get_marginal_block(parameters, self.marginal_indices)  # B, M, P

    @property
    def marginal_block_shape(self) -> Tuple[int, int]:
        return self.get_marginal_block_shape(self.marginal_indices)

    @staticmethod
    def is_marginal_block_possible(marginal_indices: MarginalIndex) -> bool:
        marginal_indices = swyft.utils.tupleize_marginal_indices(marginal_indices)
        return [len(marginal_indices[0]) == len(mi) for mi in marginal_indices]

    @classmethod
    def get_marginal_block_shape(
        cls, marginal_indices: MarginalIndex
    ) -> Tuple[int, int]:
        marginal_indices = swyft.utils.tupleize_marginal_indices(marginal_indices)
        assert cls.is_marginal_block_possible(
            marginal_indices
        ), f"Each tuple in {marginal_indices} must have the same length."
        return len(marginal_indices), len(marginal_indices[0])

    @classmethod
    def get_marginal_block(
        cls, parameters: Array, marginal_indices: MarginalIndex
    ) -> torch.Tensor:
        depth = swyft.utils.depth(marginal_indices)
        tuple_marginal_indices = swyft.utils.tupleize_marginal_indices(marginal_indices)
        assert cls.is_marginal_block_possible(
            tuple_marginal_indices
        ), f"Each tuple in {tuple_marginal_indices} must have the same length."

        if depth in [0, 1, 2]:
            return torch.stack(
                [parameters[..., mi] for mi in tuple_marginal_indices], dim=1
            )
        else:
            raise ValueError(
                f"{marginal_indices} must be of the form (a) 2, (b) [2, 3], (c) [2, [1, 3]], or (d) [[0, 1], [1, 2]]."
            )


class MarginalClassifier(nn.Module):
    def __init__(
        self,
        n_marginals: int,
        n_combined_features: int,
        hidden_features: int,
        num_blocks: int,
        dropout_probability: float = 0.0,
        use_batch_norm: bool = True,
    ) -> None:
        super().__init__()
        self.n_marginals = n_marginals
        self.n_combined_features = n_combined_features
        self.net = ResidualNetWithChannel(
            channels=self.n_marginals,
            in_features=self.n_combined_features,
            out_features=1,
            hidden_features=hidden_features,
            num_blocks=num_blocks,
            dropout_probability=dropout_probability,
            use_batch_norm=use_batch_norm,
        )

    def forward(
        self, features: torch.Tensor, marginal_block: torch.Tensor
    ) -> torch.Tensor:
        fb = features.unsqueeze(1).expand(-1, self.n_marginals, -1)  # B, M, O
        combined = torch.cat([fb, marginal_block], dim=2)  # B, M, O + P
        return self.net(combined).squeeze(-1)  # B, M


class Network(nn.Module, HeadTailClassifier):
    def __init__(
        self,
        observation_transform: nn.Module,
        parameter_transform: nn.Module,
        marginal_classifier: nn.Module,
    ) -> None:
        super().__init__()
        self.observation_transform = observation_transform
        self.parameter_transform = parameter_transform
        self.marginal_classifier = marginal_classifier

    def forward(
        self, observation: Dict[Hashable, torch.Tensor], parameters: torch.Tensor
    ) -> torch.Tensor:
        features = self.observation_transform(observation)  # B, O
        marginal_block = self.parameter_transform(parameters)  # B, M, P
        return self.marginal_classifier(features, marginal_block)  # B, M

    def head(self, observation: Dict[Hashable, torch.Tensor]) -> torch.Tensor:
        return self.observation_transform(observation)  # B, O

    def tail(self, features: torch.Tensor, parameters: torch.Tensor) -> torch.Tensor:
        marginal_block = self.parameter_transform(parameters)  # B, M, P
        return self.marginal_classifier(features, marginal_block)  # B, M


def get_marginal_classifier(
    observation_key: Hashable,
    marginal_indices: MarginalIndex,
    observation_shapes: ObsShapeType,
    n_parameters: int,
    hidden_features: int,
    num_blocks: int,
    observation_online_z_score: bool = True,
    parameter_online_z_score: bool = True,
) -> nn.Module:
    observation_transform = ObservationTransform(
        observation_key, observation_shapes, online_z_score=observation_online_z_score
    )
    n_observation_features = observation_transform.n_features

    parameter_transform = ParameterTransform(
        n_parameters, marginal_indices, online_z_score=parameter_online_z_score
    )
    n_marginals, n_block_parameters = parameter_transform.marginal_block_shape

    marginal_classifier = MarginalClassifier(
        n_marginals,
        n_observation_features + n_block_parameters,
        hidden_features=hidden_features,
        num_blocks=num_blocks,
    )

    return Network(
        observation_transform,
        parameter_transform,
        marginal_classifier,
    )


if __name__ == "__main__":
    pass



from numbers import Real
from pathlib import Path
from typing import Callable, Dict, Hashable, Mapping, Sequence, Tuple, Union

import numpy as np
import torch
from pandas import DataFrame

# If you add a type, don't forget to update autodoc_type_aliases in /docs/source/config.py

# General defintions
PathType = Union[str, Path]
Device = Union[torch.device, str]
Tensor = torch.Tensor
Array = Union[np.ndarray, torch.Tensor]
Shape = Union[torch.Size, Tuple[int, ...]]

# A list of marginals, e.g., [0, (3, 4), [3, 2]]
MarginalIndex = Union[int, Sequence[int], Sequence[Sequence[int]]]
# Strict version of that type used internally, ((0,), (2, 3), (3, 4))
StrictMarginalIndex = Tuple[Tuple[int, ...], ...]

# Map from (3, 4) --> array
MarginalToArray = Dict[Tuple[int, ...], Array]
# Map from (3, 4) --> DataFrame
MarginalToDataFrame = Dict[Tuple[int, ...], DataFrame]

ParameterNamesType = Sequence[str]
ObsType = Dict[Hashable, Array]
ObsShapeType = Mapping[Hashable, Shape]
ForwardModelType = Callable[..., ObsType]

# plotting
BaseLimitType = Tuple[Real, Real]
LimitType = Union[BaseLimitType, Sequence[BaseLimitType]]



from pathlib import Path
from typing import Sequence

import numpy as np
import torch

from swyft.types import PathType


def depth(seq: Sequence) -> int:
    if isinstance(seq, (np.ndarray, torch.Tensor)):
        return seq.ndim
    elif seq and isinstance(seq, str):
        return 0
    elif seq and isinstance(seq, Sequence):
        return 1 + max(depth(item) for item in seq)
    else:
        return 0


def is_empty(directory: PathType) -> bool:
    directory = Path(directory)
    if next(directory.iterdir(), None) is None:
        return True
    else:
        return False


def is_cuda_available() -> bool:
    return torch.cuda.is_available()


if __name__ == "__main__":
    pass



from typing import Tuple

import pytest

from swyft.types import MarginalIndex, StrictMarginalIndex
from swyft.utils.marginals import tupleize_marginal_indices
from swyft.utils.misc import depth


class TupleizeMarginals:
    marginal_indices = [
        0,
        [0, 1, 2],
        [0, [1, 2]],
        [0, 1, [2, 3]],
        [0, 1, [2, 3], [3, 4]],
        [[0, 1], [2, 3]],
        (0, (1, 2), 2),
        (0, 1),
    ]
    truth = [
        ((0,),),
        ((0,), (1,), (2,)),
        ((0,), (1, 2)),
        ((0,), (1,), (2, 3)),
        ((0,), (1,), (2, 3), (3, 4)),
        ((0, 1), (2, 3)),
        ((0,), (1, 2), (2,)),
        ((0, 1),),
    ]
    str_to_truth = {str(mi): t for mi, t in zip(marginal_indices, truth)}

    @pytest.mark.parametrize(
        "marginal_index",
        marginal_indices,
    )
    def test_compare_to_truth(self, mi: MarginalIndex) -> None:
        assert tupleize_marginal_indices(mi) == self.str_to_truth[str(mi)]

    @pytest.mark.parametrize(
        "marginal_index",
        marginal_indices,
    )
    def test_depth(mi: MarginalIndex) -> None:
        assert depth(tupleize_marginal_indices(mi)) == 2

    @pytest.mark.parametrize(
        "marginal_index",
        marginal_indices,
    )
    def test_tuple(mi: MarginalIndex) -> StrictMarginalIndex:
        mi = tupleize_marginal_indices(mi)
        assert isinstance(mi, Tuple)

    @pytest.mark.parametrize(
        "marginal_index",
        marginal_indices,
    )
    def test_nested_tuple(mi: MarginalIndex) -> StrictMarginalIndex:
        mi = tupleize_marginal_indices(mi)
        assert all(isinstance(i, Tuple) for i in mi)

    def test_sorting():
        raise NotImplementedError("Make sure the sorting is good.")


if __name__ == "__main__":
    pass



from typing import Dict, Hashable, Tuple

import torch
import torch.nn as nn


# TODO split this into a function which does the standardizing and a function which calculates the online z_scores
# That way you can easily handle the case where the user provides mean and standard deviation information
class OnlineStandardizingLayer(nn.Module):
    def __init__(
        self,
        shape: Tuple[int, ...],
        stable: bool = False,
        epsilon: float = 1e-10,
        use_average_std: bool = False,
    ) -> None:
        """Accumulate mean and variance online using the "parallel algorithm" algorithm from [1].

        Args:
            shape: shape of mean, variance, and std array. do not include batch dimension!
            stable: (optional) compute using the stable version of the algorithm [1]
            epsilon: (optional) added to the computation of the standard deviation for numerical stability.
            use_average_std: (optional) ``True`` to normalize using std averaged over the whole observation, ``False`` to normalize using std of each component of the observation.

        References:
            [1] https://en.wikipedia.org/wiki/Algorithms_for_calculating_variance#Parallel_algorithm
        """
        super().__init__()
        self.register_buffer("n", torch.tensor(0, dtype=torch.long))
        self.register_buffer("_mean", torch.zeros(shape))
        self.register_buffer("_M2", torch.zeros(shape))
        self.register_buffer("epsilon", torch.tensor(epsilon))
        self.shape = shape
        self.stable = stable
        self.use_average_std = use_average_std

    def _parallel_algorithm(
        self, x: torch.Tensor
    ) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
        assert x.shape[1:] == self.shape
        na = self.n.clone()
        nb = x.shape[0]
        nab = na + nb

        xa = self._mean.clone()
        xb = x.mean(dim=0)
        delta = xb - xa
        if self.stable:
            xab = (na * xa + nb * xb) / nab
        else:
            xab = xa + delta * nb / nab

        m2a = self._M2.clone()
        m2b = (
            x.var(dim=(0,), unbiased=False) * nb
        )  # do not use bessel's correction then multiply by total number of items in batch.
        m2ab = m2a + m2b + delta ** 2 * na * nb / nab
        return nab, xab, m2ab

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        if self.training:
            self.n, self._mean, self._M2 = self._parallel_algorithm(x)
        return (x - self.mean) / self.std

    @property
    def mean(self) -> torch.Tensor:
        return self._mean

    @property
    def var(self) -> torch.Tensor:
        if self.n > 1:
            return self._M2 / (self.n - 1)
        else:
            return torch.zeros_like(self._M2)

    @property
    def std(self) -> torch.Tensor:
        if self.use_average_std:
            return torch.sqrt(self.var + self.epsilon).mean()
        else:
            return torch.sqrt(self.var + self.epsilon)


class OnlineDictStandardizingLayer(nn.Module):
    def __init__(
        self,
        shapes: Dict[Hashable, Tuple[int, ...]],
        stable: bool = False,
        epsilon: float = 1e-10,
        use_average_std: bool = False,
    ) -> None:
        super().__init__()
        self.kwargs = dict(
            stable=stable, epsilon=epsilon, use_average_std=use_average_std
        )
        self.osls = nn.ModuleDict(
            {
                key: OnlineStandardizingLayer(shape, **self.kwargs)
                for key, shape in shapes.items()
            }
        )

    def forward(self, x: Dict[Hashable, torch.Tensor]) -> torch.Tensor:
        return {key: self.osls[key](value) for key, value in x.items()}


if __name__ == "__main__":
    pass



import decimal
import tempfile
import time
import unittest

import numpy as np
import zarr
from dask.distributed import LocalCluster, get_client

from swyft import DaskSimulator, Simulator
from swyft.prior import get_uniform_prior
from swyft.store.simulator import SimulationStatus


def model(params):
    p = np.linspace(-1, 1, 10)
    a, b = params
    mu = p * a + b
    return dict(x=mu)


def model_fail_if_negative(param):
    if param < 1:
        raise ValueError
    else:
        mu = param
    return dict(x=mu)


def model_inf_if_negative(param):
    if param < 1:
        mu = np.inf
    else:
        mu = param
    return dict(x=mu)


def _wait_for_all_tasks(timeout=20):
    client = get_client()
    start_time = time.time()
    while time.time() - start_time < timeout:
        if len(client.who_has()) == 0:
            break
        time.sleep(0.1)


class TestSimulator(unittest.TestCase):
    def test_run_simulator_all_pass(self):
        simulator = Simulator(model, sim_shapes=dict(x=(10,)), parameter_names=2)
        pars = np.random.random((100, 2))
        sims = dict(x=np.zeros((100, 10)))
        sim_status = np.full(100, SimulationStatus.RUNNING, dtype=int)

        simulator._run(
            v=pars,
            sims=sims,
            sim_status=sim_status,
            indices=np.arange(100, dtype=int),
        )

        assert np.all(sim_status == SimulationStatus.FINISHED)
        assert not np.all(np.isclose(sims["x"].sum(axis=1), 0.0))

    def test_run_simulator_fail_on_wrong_sim_shape(self):
        simulator = Simulator(model, sim_shapes=dict(x=(11,)), parameter_names=2)
        pars = np.random.random((100, 2))
        sims = dict(x=np.zeros((100, 11)))
        sim_status = np.full(100, SimulationStatus.RUNNING, dtype=int)

        self.assertRaises(
            ValueError,
            simulator._run,
            pars,
            sims,
            sim_status,
            np.arange(100, dtype=int),
        )

    def test_run_simulator_fail_on_wrong_num_params(self):
        simulator = Simulator(model, sim_shapes=dict(x=(10,)), parameter_names=1)
        pars = np.random.random((100, 1))
        sims = dict(x=np.zeros((100, 10)))
        sim_status = np.full(100, SimulationStatus.RUNNING, dtype=int)

        simulator._run(
            v=pars,
            sims=sims,
            sim_status=sim_status,
            indices=np.arange(100, dtype=int),
        )

        assert np.all(sim_status == SimulationStatus.FAILED)

    def test_run_simulator_partial_fail(self):
        simulator = Simulator(
            model_fail_if_negative, sim_shapes=dict(x=(1,)), parameter_names=1
        )
        pars = np.ones((100,))
        pars[0:49] = -1
        sims = dict(x=np.zeros((100, 1)))
        sim_status = np.full(100, SimulationStatus.RUNNING, dtype=int)

        simulator._run(
            v=pars,
            sims=sims,
            sim_status=sim_status,
            indices=np.arange(100, dtype=int),
        )

        assert np.all(sim_status[:49] == SimulationStatus.FAILED)
        assert np.all(sim_status[50:] == SimulationStatus.FINISHED)

    def test_run_simulator_partial_fail_on_inf(self):
        simulator = Simulator(
            model_inf_if_negative,
            sim_shapes=dict(x=(1,)),
            parameter_names=1,
            fail_on_non_finite=True,
        )
        pars = np.ones((100,))
        pars[0:49] = -1
        sims = dict(x=np.zeros((100, 1)))
        sim_status = np.full(100, SimulationStatus.RUNNING, dtype=int)

        simulator._run(
            v=pars,
            sims=sims,
            sim_status=sim_status,
            indices=np.arange(100, dtype=int),
        )

        assert np.all(sim_status[:49] == SimulationStatus.FAILED)
        assert np.all(sim_status[50:] == SimulationStatus.FINISHED)

    def test_run_simulator_pass_on_inf(self):
        simulator = Simulator(
            model_inf_if_negative,
            sim_shapes=dict(x=(1,)),
            parameter_names=1,
            fail_on_non_finite=False,
        )
        pars = np.ones((100,))
        pars[0:49] = -1
        sims = dict(x=np.zeros((100, 1)))
        sim_status = np.full(100, SimulationStatus.RUNNING, dtype=int)

        simulator._run(
            v=pars,
            sims=sims,
            sim_status=sim_status,
            indices=np.arange(100, dtype=int),
        )

        assert np.all(sim_status[50:] == SimulationStatus.FINISHED)

    def test_run_simulator_with_zarr_memory_store(self):
        """Test the simulator with a store based on Zarr MemoryStore."""
        simulator = Simulator(model, sim_shapes=dict(x=(10,)), parameter_names=2)

        pars = zarr.zeros((100, 2))
        pars[:, :] = np.random.random(pars.shape)
        x = zarr.zeros((100, 10))
        sims = dict(x=x.oindex)
        sim_status = zarr.full(100, SimulationStatus.RUNNING, dtype="int")

        simulator._run(
            v=pars,
            sims=sims,
            sim_status=sim_status.oindex,
            indices=np.arange(100, dtype=int),
        )

        assert np.all(sim_status[:] == SimulationStatus.FINISHED)
        assert not np.all(np.isclose(sims["x"][:, :].sum(axis=1), 0.0))

    def test_run_simulator_with_zarr_directory_store(self):
        """Test simulator with a store based on Zarr DirectoryStore."""
        simulator = Simulator(model, sim_shapes=dict(x=(10,)), parameter_names=2)

        with tempfile.TemporaryDirectory() as tmpdir:
            pars = zarr.open(f"{tmpdir}/pars.zarr", shape=(100, 2))
            pars[:, :] = np.random.random(pars.shape)
            x = zarr.open(f"{tmpdir}/x.zarr", shape=(100, 10))
            x[:, :] = 0.0
            sims = dict(x=x.oindex)
            sim_status = zarr.open(f"{tmpdir}/sim_status.zarr", shape=(100,))
            sim_status[:] = np.full(100, SimulationStatus.RUNNING, dtype="int")

            simulator._run(
                v=pars,
                sims=sims,
                sim_status=sim_status.oindex,
                indices=np.arange(100, dtype=int),
            )

            assert np.all(sim_status[:] == SimulationStatus.FINISHED)
            assert not np.all(np.isclose(sims["x"][:, :].sum(axis=1), 0.0))

    def test_setup_simulator_from_model_function(self):
        dimension = 2
        prior = get_uniform_prior(-1 * np.ones(dimension), np.ones(dimension))
        simulator = Simulator.from_model(model, prior)
        assert len(simulator.parameter_names) == 2
        assert simulator.sim_shapes == {"x": (10,)}

    def test_run_a_simulator_that_is_setup_from_command_line(self):
        def set_input(v):
            ctx = decimal.Context()
            ctx.prec = 16
            return "{} + {}\n".format(
                ctx.create_decimal(repr(v[0])), ctx.create_decimal(repr(v[1]))
            )

        def get_output(stdout, _):
            return {"sum": float(stdout)}

        simulator = Simulator.from_command(
            command="bc -l",
            set_input_method=set_input,
            get_output_method=get_output,
            parameter_names=2,
            sim_shapes={"sum": ()},
        )

        pars = np.random.random((100, 2))
        sims = dict(sum=np.zeros(100))
        sim_status = np.full(100, SimulationStatus.RUNNING, dtype=int)

        simulator._run(
            v=pars,
            sims=sims,
            sim_status=sim_status,
            indices=np.arange(100, dtype=int),
        )

        assert np.all(sim_status == SimulationStatus.FINISHED)
        assert not np.any(np.isclose(sims["sum"], 0.0))


class TestDaskSimulator(unittest.TestCase):
    def test_connect_to_dask_simulator(self):
        with tempfile.TemporaryDirectory() as tmpdir:
            with LocalCluster(
                n_workers=2,
                processes=True,
                threads_per_worker=1,
                local_directory=tmpdir,
            ) as cluster:
                simulator = DaskSimulator(
                    model_fail_if_negative, sim_shapes=dict(x=(10,)), parameter_names=2
                )
                simulator.set_dask_cluster(cluster)
                assert cluster.name == simulator.client.cluster.name

    def test_run_simulator_with_processes_and_numpy_array(self):
        """
        If the store is in memory (here a Numpy array) and the Dask workers do not
        share  memory with the client (e.g. we have a processes-based cluster),
        collect_in_memory must be set to True.
        """
        with tempfile.TemporaryDirectory() as tmpdir:
            with LocalCluster(
                n_workers=2,
                processes=True,
                threads_per_worker=1,
                local_directory=tmpdir,
            ) as cluster:
                simulator = DaskSimulator(
                    model, sim_shapes=dict(x=(10,)), parameter_names=2
                )
                simulator.set_dask_cluster(cluster)

                pars = np.random.random((100, 2))
                sims = dict(x=np.zeros((100, 10)))
                sim_status = np.full(100, SimulationStatus.RUNNING, dtype=int)

                simulator._run(
                    v=pars,
                    sims=sims,
                    sim_status=sim_status,
                    indices=np.arange(100, dtype=int),
                    collect_in_memory=True,
                    batch_size=20,
                )

                assert np.all(sim_status == SimulationStatus.FINISHED)
                assert not np.all(np.isclose(sims["x"].sum(axis=1), 0.0))

    def test_run_simulator_with_processes_and_zarr_memory_store(self):
        """
        If the store is in memory (here a Zarr MemoryStore) and the Dask workers do
        not share memory with the client (i.e. we have a processes-based cluster),
        collect_in_memory must be set to True.
        """
        with tempfile.TemporaryDirectory() as tmpdir:
            with LocalCluster(
                n_workers=2,
                processes=True,
                threads_per_worker=1,
                local_directory=tmpdir,
            ) as cluster:
                simulator = DaskSimulator(
                    model, sim_shapes=dict(x=(10,)), parameter_names=2
                )
                simulator.set_dask_cluster(cluster)

                pars = zarr.zeros((100, 2))
                pars[:, :] = np.random.random(pars.shape)
                x = zarr.zeros((100, 10))
                sims = dict(x=x.oindex)
                sim_status = zarr.full(100, SimulationStatus.RUNNING, dtype="int")

                simulator._run(
                    v=pars,
                    sims=sims,
                    sim_status=sim_status.oindex,
                    indices=np.arange(100, dtype=int),
                    collect_in_memory=True,
                    batch_size=20,
                )

                assert np.all(sim_status[:] == SimulationStatus.FINISHED)
                assert not np.all(np.isclose(sims["x"][:, :].sum(axis=1), 0.0))

    def test_run_simulator_with_threads_and_zarr_directory_store(self):
        """
        If the store is on disk (here a Zarr DirectoryStore), collect_in_memory can
        be set to False (but synchronization needs to be employed).
        """
        with tempfile.TemporaryDirectory() as tmpdir:
            with LocalCluster(
                n_workers=2,
                processes=False,
                threads_per_worker=1,
                local_directory=tmpdir,
            ) as cluster:
                simulator = DaskSimulator(
                    model, sim_shapes=dict(x=(10,)), parameter_names=2
                )
                simulator.set_dask_cluster(cluster)
                pars = zarr.open(f"{tmpdir}/pars.zarr", shape=(100, 2))
                pars[:, :] = np.random.random(pars.shape)
                x = zarr.open(
                    f"{tmpdir}/x.zarr",
                    shape=(100, 10),
                    synchronizer=zarr.ThreadSynchronizer(),
                )
                x[:, :] = 0.0
                sims = dict(x=x.oindex)
                sim_status = zarr.open(
                    f"{tmpdir}/sim_status.zarr",
                    shape=(100,),
                    synchronizer=zarr.ThreadSynchronizer(),
                )
                sim_status[:] = np.full(100, SimulationStatus.RUNNING, dtype="int")

                # the following is non-blocking (it immediately returns)
                simulator._run(
                    v=pars,
                    sims=sims,
                    sim_status=sim_status.oindex,
                    indices=np.arange(100, dtype=int),
                    collect_in_memory=False,
                    batch_size=20,
                )

                # need to wait for tasks to be completed
                _wait_for_all_tasks()

                assert np.all(sim_status[:] == SimulationStatus.FINISHED)
                assert not np.all(np.isclose(sims["x"][:, :].sum(axis=1), 0.0))

    def test_run_simulator_with_processes_and_zarr_directory_store(self):
        """
        If the store is on disk (here a Zarr DirectoryStore), collect_in_memory can
        be set to False (but synchronization needs to be employed).
        """
        with tempfile.TemporaryDirectory() as tmpdir:
            with LocalCluster(
                n_workers=2,
                processes=True,
                threads_per_worker=1,
                local_directory=tmpdir,
            ) as cluster:
                simulator = DaskSimulator(
                    model, sim_shapes=dict(x=(10,)), parameter_names=2
                )
                simulator.set_dask_cluster(cluster)
                pars = zarr.open(f"{tmpdir}/pars.zarr", shape=(100, 2))
                pars[:, :] = np.random.random(pars.shape)
                synchronizer = zarr.ProcessSynchronizer(path=f"{tmpdir}/x.sync")
                x = zarr.open(
                    f"{tmpdir}/x.zarr", shape=(100, 10), synchronizer=synchronizer
                )
                x[:, :] = 0.0
                sims = dict(x=x.oindex)
                synchronizer = zarr.ProcessSynchronizer(
                    path=f"{tmpdir}/sim_status.sync"
                )
                sim_status = zarr.open(
                    f"{tmpdir}/sim_status.zarr",
                    shape=(100,),
                    synchronizer=synchronizer,
                    dtype="int",
                )
                sim_status[:] = np.full(100, SimulationStatus.RUNNING, dtype="int")

                # the following is non-blocking (it immediately returns)
                simulator._run(
                    v=pars,
                    sims=sims,
                    sim_status=sim_status.oindex,
                    indices=np.arange(100, dtype=int),
                    collect_in_memory=False,
                    batch_size=20,
                )

                # need to wait for tasks to be completed
                _wait_for_all_tasks()

                assert np.all(sim_status[:] == SimulationStatus.FINISHED)
                assert not np.all(np.isclose(sims["x"][:, :].sum(axis=1), 0.0))

    def test_run_a_simulator_that_is_setup_from_command_line(self):
        """Run a simulator based on a command line model

        Notes:
            Need to have a cluster which uses only processes since each
            instance of the model is run in a separate subdirectory.
        """
        with tempfile.TemporaryDirectory() as tmpdir:
            with LocalCluster(
                n_workers=2,
                processes=True,
                threads_per_worker=1,
                local_directory=tmpdir,
            ) as cluster:

                def set_input(v):
                    ctx = decimal.Context()
                    ctx.prec = 16
                    return "{} + {}\n".format(
                        ctx.create_decimal(repr(v[0])), ctx.create_decimal(repr(v[1]))
                    )

                def get_output(stdout, _):
                    return {"sum": float(stdout)}

                simulator = DaskSimulator.from_command(
                    command="bc -l",
                    set_input_method=set_input,
                    get_output_method=get_output,
                    parameter_names=2,
                    sim_shapes={"sum": ()},
                )
                simulator.set_dask_cluster(cluster)

                pars = np.random.randint(1, 100, size=(100, 2))
                sims = dict(sum=np.zeros(100))
                sim_status = np.full(100, SimulationStatus.RUNNING, dtype=int)

                simulator._run(
                    v=pars,
                    sims=sims,
                    sim_status=sim_status,
                    indices=np.arange(100, dtype=int),
                )

                assert np.all(sim_status == SimulationStatus.FINISHED)
                assert not np.any(np.isclose(sims["sum"], 0.0))



from typing import Dict, Hashable, Tuple

import torch
import torch.nn as nn

import swyft
import swyft.utils
from swyft.networks.linear import LinearWithChannel
from swyft.types import Array, MarginalIndex


def is_marginal_block_possible(marginal_indices: MarginalIndex) -> bool:
    marginal_indices = swyft.utils.tupleize_marginals(marginal_indices)
    return [len(marginal_indices[0]) == len(mi) for mi in marginal_indices]


def get_marginal_block_shape(marginal_indices: MarginalIndex) -> Tuple[int, int]:
    marginal_indices = swyft.utils.tupleize_marginals(marginal_indices)
    assert is_marginal_block_possible(
        marginal_indices
    ), f"Each tuple in {marginal_indices=} must have the same length."
    return len(marginal_indices), len(marginal_indices[0])


def get_marginal_block(parameters: Array, marginal_indices: MarginalIndex) -> Array:
    depth = swyft.utils.depth(marginal_indices)
    tuple_marginal_indices = swyft.utils.tupleize_marginals(marginal_indices)
    assert is_marginal_block_possible(
        tuple_marginal_indices
    ), f"Each tuple in {tuple_marginal_indices=} must have the same length."

    if depth in [0, 1, 2]:
        return torch.stack(
            [parameters[..., mi] for mi in tuple_marginal_indices], dim=1
        )
    else:
        raise ValueError(
            f"{marginal_indices=} must be of the form (a) 2, (b) [2, 3], (c) [2, [1, 3]], or (d) [[0, 1], [1, 2]]."
        )


class Head(nn.Module):
    def __init__(
        self, observation_key: Hashable, in_features: int, out_features: int
    ) -> None:
        super().__init__()
        self.observation_key = observation_key
        self.in_features = in_features
        self.out_features = out_features
        self.net = nn.Linear(in_features, out_features)

    def forward(self, observation: Dict[Hashable, torch.Tensor]) -> torch.Tensor:
        x = observation[self.observation_key]
        return self.net(x)


class Combinator(nn.Module):
    def __init__(self, head_features: int, marginal_indices: MarginalIndex) -> None:
        super().__init__()
        self.head_features = head_features
        self.marginal_indices = marginal_indices
        self.n_marginals, self.n_parameters = get_marginal_block_shape(
            self.marginal_indices
        )

    def forward(self, features: torch.Tensor, parameters: torch.Tensor) -> torch.Tensor:
        mb = get_marginal_block(parameters, self.marginal_indices)  # B, M, P
        fb = features.unsqueeze(1).expand(-1, self.n_marginals, -1)  # B, M, F
        return torch.cat([fb, mb], dim=2)  # B, M, F + P


class Tail(nn.Module):
    def __init__(self, n_marginals: int, n_combined_features: int) -> None:
        super().__init__()
        self.n_marginals = n_marginals
        self.n_combined_features = n_combined_features
        self.net = LinearWithChannel(self.n_marginals, self.n_combined_features, 1)

    def forward(self, combined_features) -> torch.Tensor:
        print(combined_features.shape)
        return self.net(combined_features).squeeze(-1)


class MarginalLogRatioEstimator(nn.Module):
    def __init__(self, head: nn.Module, combinator: nn.Module, tail: nn.Module) -> None:
        super().__init__()
        self.head = head
        self.combinator = combinator
        self.tail = tail

    def forward(
        self, observation: Dict[Hashable, torch.Tensor], parameters: torch.Tensor
    ) -> torch.Tensor:
        h_out = self.head(observation)
        c_out = self.combinator(h_out, parameters)
        return self.tail(c_out)


def main():
    b, o = 10, 7
    f = 10
    p = 5
    marginal_indices = [1, 2, 3]
    key = "x"
    x = {key: torch.rand(b, o)}
    theta = torch.rand(b, p)

    nm, nbp = get_marginal_block_shape(marginal_indices)

    mre = MarginalLogRatioEstimator(
        Head(key, o, f),
        Combinator(f, marginal_indices),
        Tail(nm, f + nbp),
    )
    out = mre(x, theta)
    breakpoint()


if __name__ == "__main__":
    main()



import tempfile
from itertools import chain, product
from pathlib import Path
from typing import Callable, Tuple

import numpy as np
import pytest
import torch
from scipy import stats
from toolz import compose
from torch.distributions import Normal, Uniform

from swyft.prior import InterpolatedTabulatedDistribution, Prior
from swyft.utils import array_to_tensor, tensor_to_array

dimensions = [1, 2, 5]
uniform_hyperparameters = [(torch.zeros(d), torch.ones(d)) for d in dimensions]
uniform_hyperparameters += [
    (-2 * torch.ones(d), -1 * torch.ones(d)) for d in dimensions
]
uniform_hyperparameters += [(-1 * torch.ones(d), 1 * torch.ones(d)) for d in dimensions]
uniform_hyperparameters += [(1 * torch.ones(d), 2 * torch.ones(d)) for d in dimensions]

normal_hyperparameters = [(torch.zeros(d), torch.ones(d)) for d in dimensions]
normal_hyperparameters += [(5 * torch.ones(d), torch.ones(d)) for d in dimensions]
normal_hyperparameters += [
    (5 * torch.randn(d), 0.1 * torch.ones(d)) for d in dimensions
]


class TestInterpolatedTabulatedDistribution:
    def setup_method(self, method):
        torch.manual_seed(0)
        self.n = 1_000
        self.n_grid_points = 10_000

    @pytest.mark.parametrize(
        "distribution, args",
        chain(
            product([Uniform], uniform_hyperparameters),
            product([Normal], normal_hyperparameters),
        ),
    )
    def test_u(self, distribution: Callable, args: Tuple[torch.Tensor, ...]) -> None:
        distribution = distribution(*args)
        samples = distribution.sample((self.n,))
        hypercube_samples = distribution.cdf(samples).numpy()
        _, n_parameters = samples.shape

        itd = InterpolatedTabulatedDistribution(
            compose(tensor_to_array, distribution.icdf, array_to_tensor),
            n_parameters,
            n_grid_points=self.n_grid_points,
        )

        hypercube_samples_itd = itd.cdf(samples)
        assert np.allclose(
            hypercube_samples, hypercube_samples_itd, atol=1e-4, rtol=5e-3
        )

    @pytest.mark.skip(reason="This is an inaccurate feature and should be removed.")
    @pytest.mark.parametrize(
        "distribution, args",
        chain(
            product([Uniform], uniform_hyperparameters),
            product([Normal], normal_hyperparameters),
        ),
    )
    def test_v(self, distribution: Callable, args: Tuple[torch.Tensor, ...]) -> None:
        distribution = distribution(*args)
        samples = distribution.sample((self.n,))
        hypercube_samples = distribution.cdf(samples).numpy()
        _, n_parameters = samples.shape

        itd = InterpolatedTabulatedDistribution(
            compose(tensor_to_array, distribution.icdf, array_to_tensor),
            n_parameters,
            n_grid_points=self.n_grid_points,
        )

        samples_itd = itd.icdf(hypercube_samples)

        if isinstance(distribution, Uniform):
            assert np.allclose(samples, samples_itd)
        elif isinstance(distribution, Normal):
            assert np.allclose(samples, samples_itd, atol=1e-1, rtol=1e-1)


class TestPrior:
    def setup_method(self, method):
        torch.manual_seed(0)
        self.n = 10_000

    @pytest.mark.parametrize(
        "distribution, args",
        chain(
            product([Uniform], uniform_hyperparameters),
            product([Normal], normal_hyperparameters),
        ),
    )
    def test_u_from_torch_distribution(
        self, distribution: Callable, args: Tuple[torch.Tensor, ...]
    ) -> None:
        distribution = distribution(*args)
        samples = distribution.sample((self.n,))
        hypercube_samples_true = distribution.cdf(samples).numpy()

        prior = Prior.from_torch_distribution(distribution)
        hypercube_samples_esti = prior.cdf(samples)
        assert np.allclose(hypercube_samples_true, hypercube_samples_esti)

    @pytest.mark.parametrize(
        "distribution, args",
        chain(
            product([Uniform], uniform_hyperparameters),
            product([Normal], normal_hyperparameters),
        ),
    )
    def test_v_from_torch_distribution(
        self, distribution: Callable, args: Tuple[torch.Tensor, ...]
    ) -> None:
        distribution = distribution(*args)
        samples_true = distribution.sample((self.n,))
        hypercube_samples = distribution.cdf(samples_true).numpy()

        prior = Prior.from_torch_distribution(distribution)
        samples_esti = prior.icdf(hypercube_samples)

        if isinstance(distribution, Uniform):
            assert np.allclose(samples_true, samples_esti)
        elif isinstance(distribution, Normal):
            assert np.allclose(samples_true, samples_esti, atol=1e-4, rtol=5e-3)


class TestSaveLoadPrior:
    @classmethod
    def setup_class(cls):
        cls.directory = tempfile.TemporaryDirectory()

    @classmethod
    def teardown_class(cls):
        cls.directory.cleanup()

    def test_save_load_init(self):
        normal = stats.norm(0, 1)
        prior = Prior(
            cdf=normal.cdf,
            icdf=normal.ppf,
            log_prob=normal.logpdf,
            n_parameters=1,
        )

        # Saving
        path = Path(self.directory.name) / f"from_init_scipy"
        prior.save(path)

        # Loading
        prior_loaded = Prior.load(path)

        # Testing by cdf
        # (icdf or log_prob would also be fine.)
        samples = np.random.randn(1_000)
        cdf_true = prior.log_prob(samples)
        cdf_esti = prior_loaded.log_prob(samples)
        assert np.allclose(cdf_true, cdf_esti)

    @pytest.mark.parametrize(
        "distribution, args",
        [
            (Uniform, uniform_hyperparameters[0]),
            (Normal, normal_hyperparameters[0]),
        ],
    )
    def test_save_load_from_torch_distribution(self, distribution, args):
        distribution = distribution(*args)
        prior = Prior.from_torch_distribution(distribution)

        # Saving
        path = (
            Path(self.directory.name)
            / f"from_torch_distribution_{distribution.__class__.__name__}"
        )
        prior.save(path)

        # Loading
        prior_loaded = Prior.load(path)

        # Testing by cdf
        # (icdf or log_prob would also be fine.)
        samples = prior.distribution.sample((1_000,))
        cdf_true = prior.cdf(samples)
        cdf_esti = prior_loaded.cdf(samples)
        assert np.allclose(cdf_true, cdf_esti)


if __name__ == "__main__":
    pass



import math
from typing import Callable

import torch
import torch.nn as nn
from torch.nn import functional as F
from torch.nn import init


# Inspired by: https://github.com/pytorch/pytorch/issues/36591
class LinearWithChannel(torch.nn.Module):
    def __init__(self, channels: int, in_features: int, out_features: int) -> None:
        super(LinearWithChannel, self).__init__()
        self.weights = torch.nn.Parameter(
            torch.empty((channels, out_features, in_features))
        )
        self.bias = torch.nn.Parameter(torch.empty(channels, out_features))

        # Initialize weights
        torch.nn.init.kaiming_uniform_(self.weights, a=math.sqrt(5))
        fan_in, _ = torch.nn.init._calculate_fan_in_and_fan_out(self.weights)
        bound = 1 / math.sqrt(fan_in) if fan_in > 0 else 0
        torch.nn.init.uniform_(self.bias, -bound, bound)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        assert x.ndim >= 2, "Requires (..., channel, features) shape."
        x = x.unsqueeze(-1)
        result = torch.matmul(self.weights, x).squeeze(-1) + self.bias
        return result


class BatchNorm1dWithChannel(nn.BatchNorm1d):
    def __init__(
        self,
        num_channels: int,
        num_features: int,
        eps: float = 1e-5,
        momentum: float = 0.1,
        affine: bool = True,
        track_running_stats: bool = True,
    ) -> None:
        """BatchNorm1d over the batch, N. Requires shape (N, C, L).

        Otherwise, same as torch.nn.BatchNorm1d with extra num_channel. Cannot do the temporal batch norm case.
        """
        num_features = num_channels * num_features
        super().__init__(num_features, eps, momentum, affine, track_running_stats)
        self.flatten = nn.Flatten()

    def forward(self, input: torch.Tensor) -> torch.Tensor:
        n, c, f = input.shape
        flat = self.flatten(input)
        batch_normed = super().forward(flat)
        return batch_normed.reshape(n, c, f)


# inspired by https://github.com/bayesiains/nflows/blob/master/nflows/nn/nets/resnet.py
class ResidualBlockWithChannel(nn.Module):
    """A general-purpose residual block. Works only with channelized 1-dim inputs."""

    def __init__(
        self,
        channels: int,
        features: int,
        activation: Callable = F.relu,
        dropout_probability: float = 0.0,
        use_batch_norm: bool = False,
        zero_initialization: bool = True,
    ) -> None:
        super().__init__()
        self.activation = activation

        self.use_batch_norm = use_batch_norm
        if use_batch_norm:
            self.batch_norm_layers = nn.ModuleList(
                [BatchNorm1dWithChannel(channels, features, eps=1e-3) for _ in range(2)]
            )
        self.linear_layers = nn.ModuleList(
            [LinearWithChannel(channels, features, features) for _ in range(2)]
        )
        self.dropout = nn.Dropout(p=dropout_probability)
        if zero_initialization:
            init.uniform_(self.linear_layers[-1].weights, -1e-3, 1e-3)
            init.uniform_(self.linear_layers[-1].bias, -1e-3, 1e-3)

    def forward(self, inputs: torch.Tensor) -> torch.Tensor:
        temps = inputs
        if self.use_batch_norm:
            temps = self.batch_norm_layers[0](temps)
        temps = self.activation(temps)
        temps = self.linear_layers[0](temps)
        if self.use_batch_norm:
            temps = self.batch_norm_layers[1](temps)
        temps = self.activation(temps)
        temps = self.dropout(temps)
        temps = self.linear_layers[1](temps)
        return inputs + temps


# inspired by https://github.com/bayesiains/nflows/blob/master/nflows/nn/nets/resnet.py
class ResidualNetWithChannel(nn.Module):
    """A general-purpose residual network. Works only with channelized 1-dim inputs."""

    def __init__(
        self,
        channels: int,
        in_features: int,
        out_features: int,
        hidden_features: int,
        num_blocks: int = 2,
        activation: Callable = F.relu,
        dropout_probability: float = 0.0,
        use_batch_norm: bool = False,
    ) -> None:
        super().__init__()
        self.hidden_features = hidden_features
        self.initial_layer = LinearWithChannel(channels, in_features, hidden_features)
        self.blocks = nn.ModuleList(
            [
                ResidualBlockWithChannel(
                    channels=channels,
                    features=hidden_features,
                    activation=activation,
                    dropout_probability=dropout_probability,
                    use_batch_norm=use_batch_norm,
                )
                for _ in range(num_blocks)
            ]
        )
        self.final_layer = LinearWithChannel(channels, hidden_features, out_features)

    def forward(self, inputs: torch.Tensor) -> torch.Tensor:
        temps = self.initial_layer(inputs)
        for block in self.blocks:
            temps = block(temps)
        outputs = self.final_layer(temps)
        return outputs



import numpy as np
import pytest
from matplotlib.figure import Figure

from swyft.plot.mass import empirical_z_score_corner
from swyft.utils.marginals import get_corner_marginal_indices


class TestMass:
    @pytest.mark.parametrize("n_parameters", [2, 3])
    def test_empirical_z_score_corner(self, n_parameters: int) -> None:
        n_samples = 1_000
        marginal_indices_1d, marginal_indices_2d = get_corner_marginal_indices(
            n_parameters
        )
        empirical_mass_1d = {
            marginal_index: np.random.rand(n_samples)
            for marginal_index in marginal_indices_1d
        }
        empirical_mass_2d = {
            marginal_index: np.random.rand(n_samples)
            for marginal_index in marginal_indices_2d
        }

        fig, axes = empirical_z_score_corner(empirical_mass_1d, empirical_mass_2d)
        assert isinstance(fig, Figure)
        assert axes.shape == (n_parameters, n_parameters)



from dataclasses import dataclass
from typing import Tuple, TypeVar

import numpy as np
import pandas as pd

from swyft.types import (
    MarginalIndex,
    MarginalToArray,
    MarginalToDataFrame,
    StrictMarginalIndex,
)
from swyft.utils.marginals import filter_marginals_by_dim, tupleize_marginal_indices

WeightedMarginalSamplesType = TypeVar(
    "WeightedMarginalSamplesType", bound="WeightedMarginalSamples"
)


@dataclass
class WeightedMarginalSamples:
    weights: MarginalToArray
    v: np.ndarray

    @staticmethod
    def _select_marginal_index(marginal_index: MarginalIndex) -> Tuple[int, ...]:
        marginal_index = tupleize_marginal_indices(marginal_index)
        assert (
            len(marginal_index) == 1
        ), "weighted marginal samples can only be recovered one index at a time"
        return marginal_index[0]

    def get_logweight(self, marginal_index: MarginalIndex) -> np.ndarray:
        """access the logweight for a certain marginal by marginal_index

        Args:
            marginal_index: which marginal to select. one at at time.

        Returns:
            logweight
        """
        marginal_index = self._select_marginal_index(marginal_index)
        logweight = self.weights[marginal_index]
        return logweight

    def get_logweight_marginal(
        self, marginal_index: MarginalIndex
    ) -> Tuple[np.ndarray, np.ndarray]:
        """access the logweight and parameter values for a marginal by index

        Args:
            marginal_index: which marginal to select. one at a time.

        Returns:
            logweight, marginal: the logweight and the parameter values
        """
        marginal_index = self._select_marginal_index(marginal_index)
        logweight = self.get_logweight(marginal_index)
        marginal = self.v[:, marginal_index]
        return logweight, marginal

    def get_df(self, marginal_index: MarginalIndex) -> pd.DataFrame:
        """convert a weighted marginal into a dataframe with the marginal_indices, 'weight', and 'logweight' as columns

        Args:
            marginal_index: which marginal to select. one at a time.

        Returns:
            DataFrame with marginal_indices, 'weight', and 'logweight' for columns
        """
        marginal_index = self._select_marginal_index(marginal_index)
        logweight, marginal = self.get_logweight_marginal(marginal_index)
        weight = np.exp(logweight)

        data = np.concatenate(
            [marginal, weight[..., None], logweight[..., None]], axis=-1
        )
        columns = list(marginal_index) + ["weight"] + ["logweight"]
        return pd.DataFrame(data=data, columns=columns)

    def get_df_dict(self) -> MarginalToDataFrame:
        """produce a map from marginal_index to df for all dfs and marginal_indices"""
        return {
            marginal_index: self.get_df(marginal_index)
            for marginal_index in self.marginal_indices
        }

    def filter_by_dim(self, dim: int) -> WeightedMarginalSamplesType:
        weights = filter_marginals_by_dim(self.weights, dim)
        return self.__class__(weights, self.v)

    @property
    def marginal_indices(self) -> StrictMarginalIndex:
        return tuple(self.weights.keys())



from typing import Optional, Sequence, Tuple, Union

import matplotlib.pyplot as plt
import numpy as np
import seaborn as sns
from matplotlib.figure import Figure
from pandas import DataFrame
from torch import Tensor
from tqdm import tqdm

from swyft.types import Array, LimitType, MarginalToArray, MarginalToDataFrame
from swyft.utils.marginals import get_d_dim_marginal_indices, get_df_dict_from_marginals
from swyft.weightedmarginals import WeightedMarginalSamples


def split_corner_axes(axes: np.ndarray) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    diag = np.diag(axes)
    lower = axes[np.tril(axes, -1).nonzero()]
    upper = axes[np.triu(axes, 1).nonzero()]
    return lower, diag, upper


def _set_weight_keyword(df: DataFrame) -> Optional[str]:
    if "weight" in df.columns:
        return "weight"
    elif "weights" in df.columns:
        return "weights"
    else:
        return None


def _marginaldict_weightedmarginaldict_to_dfdict(
    marginal: Union[WeightedMarginalSamples, MarginalToArray]
) -> MarginalToDataFrame:
    """convert marginal samples, marginal weighted samples, and marginal dataframes to marginal dataframes.

    Args:
        marginal: input to convert

    Raises:
        TypeError: when not one of the supported types

    Returns:
        MarginalToDataFrame: a dictionary mapping marginal to dataframe
    """
    if isinstance(marginal, WeightedMarginalSamples):
        return marginal.get_df_dict()
    elif isinstance(list(marginal.values())[0], (np.ndarray, Tensor)):
        return get_df_dict_from_marginals(marginal)
    elif isinstance(list(marginal.values())[0], DataFrame):
        return marginal
    else:
        raise TypeError(
            "Only marginal samples, marginal weighted samples, and marginal dataframes are supported."
        )


def hist1d(
    marginal_1d: Union[WeightedMarginalSamples, MarginalToArray, MarginalToDataFrame],
    figsize: Optional[Tuple[float, float]] = None,
    bins: int = 100,
    kde: Optional[bool] = False,
    xlim: LimitType = None,
    ylim: LimitType = None,
    truth: Array = None,
    labels: Sequence[str] = None,
    ticks: bool = True,
    ticklabels: bool = True,
    ticklabelsize: str = "x-small",
    tickswhich: str = "both",
    labelrotation: float = 45.0,
    space_between_axes: float = 0.1,
) -> Tuple[Figure, np.ndarray]:
    """create a row of 1d histograms from marginals

    Args:
        marginal_1d: one dimensional marginal samples, weighted samples, or dataframes
        figsize: choose the figsize. like matplotlib.
        bins: number of bins for the histogram
        kde: do a kernel density estimate to produce isocontour lines? (may be expensive)
        xlim: set the xlim. either a single tuple for the same value on all plots, or a sequence of tuples for every plot.
        ylim: set the ylim. either a single tuple for the same value on all plots, or a sequence of tuples for every plot.
        truth: array denoting the true parameter which generated the observation.
        labels: the string labels for the parameters.
        ticks: whether to show ticks
        ticklabels: whether to show the value of the ticks. only functions when `ticks=True`.
        ticklabelsize: set size of tick labels. see `plt.tick_params`
        tickswhich: whether to affect major or minor ticks. see `plt.tick_params`
        labelrotation: tick label rotation. see `plt.tick_params`
        space_between_axes: changes the `wspace` and `hspace` between subplots. see `plt.subplots_adjust`

    Returns:
        matplotlib figure, np array of matplotlib axes
    """
    marginal_1d = _marginaldict_weightedmarginaldict_to_dfdict(marginal_1d)
    d = len(marginal_1d)

    fig, axes = plt.subplots(ncols=d, sharey=True, figsize=figsize)

    lb = 0.125
    tr = 0.9
    fig.subplots_adjust(
        left=lb,
        bottom=lb,
        right=tr,
        top=tr,
        wspace=space_between_axes,
        hspace=space_between_axes,
    )

    for i, (k, ax) in enumerate(zip(sorted(marginal_1d.keys()), axes.flatten())):
        df = marginal_1d[k]
        sns.histplot(
            data=df,
            x=k[0],
            stat="density",
            weights=_set_weight_keyword(df),
            bins=bins,
            ax=ax,
            element="step",
            fill=False,
            color="k",
        )
        if kde:
            sns.kdeplot(
                data=df,
                x=k[0],
                weights=_set_weight_keyword(df),
                ax=ax,
                fill=False,
                color="b",
            )
        ax.tick_params(
            axis="x",
            which=tickswhich,
            bottom=ticks,
            direction="out",
            labelbottom=ticks and ticklabels,
            labelrotation=labelrotation,
            labelsize=ticklabelsize,
        )

        if labels is not None:
            ax.set_xlabel(labels[i])

        if truth is not None:
            ax.axvline(truth[i], color="r")

        if xlim is None:
            pass
        elif isinstance(xlim[0], (int, float)):
            ax.set_xlim(*xlim)
        elif isinstance(xlim[0], (tuple, list)):
            ax.set_xlim(*xlim[i])
        else:
            raise NotImplementedError("xlim should be a tuple or a list of tuples.")

        if ylim is None:
            pass
        elif isinstance(ylim[0], (int, float)):
            ax.set_ylim(*ylim)
        elif isinstance(ylim[0], (tuple, list)):
            ax.set_ylim(*ylim[i])
        else:
            raise NotImplementedError("ylim should be a tuple or a list of tuples.")

    fig.align_labels()
    return fig, axes


def corner(
    marginal_1d: Union[WeightedMarginalSamples, MarginalToArray, MarginalToDataFrame],
    marginal_2d: Union[WeightedMarginalSamples, MarginalToArray, MarginalToDataFrame],
    figsize: Optional[Tuple[float, float]] = None,
    bins: int = 100,
    kde: Optional[bool] = False,
    xlim: LimitType = None,
    ylim_lower: LimitType = None,
    truth: Array = None,
    levels: int = 3,
    labels: Sequence[str] = None,
    ticks: bool = True,
    ticklabels: bool = True,
    ticklabelsize: str = "x-small",
    tickswhich: str = "both",
    labelrotation: float = 45.0,
    space_between_axes: float = 0.1,
) -> Tuple[Figure, np.ndarray]:
    """create a corner plot from a dictionary of marginals

    Args:
        marginal_1d: one dimensional marginal samples, weighted samples, or dataframes
        marginal_2d: two dimensional marginal samples, weighted samples, or dataframes
        figsize: choose the figsize. like matplotlib.
        bins: number of bins for the histogram
        kde: do a kernel density estimate to produce isocontour lines? (may be expensive)
        xlim: set the xlim. either a single tuple for the same value on all plots, or a sequence of tuples for every column in the plot.
        ylim_lower: set the ylim for the 2d histograms on the lower triangle. either a single tuple for the same value on all plots, or a sequence of tuples for every row in the plot. first row is not considered.
        truth: array denoting the true parameter which generated the observation.
        levels: number of isocontour lines to plot. only functions when `kde=True`.
        labels: the string labels for the parameters.
        ticks: whether to show ticks on the bottom row and leftmost column
        ticklabels: whether to show the value of the ticks bottom row and leftmost column. only functions when `ticks=True`.
        ticklabelsize: set size of tick labels. see `plt.tick_params`
        tickswhich: whether to affect major or minor ticks. see `plt.tick_params`
        labelrotation: tick label rotation. see `plt.tick_params`
        space_between_axes: changes the `wspace` and `hspace` between subplots. see `plt.subplots_adjust`

    Returns:
        matplotlib figure, np array of matplotlib axes
    """
    marginals_1d = _marginaldict_weightedmarginaldict_to_dfdict(marginal_1d)
    marginals_2d = _marginaldict_weightedmarginaldict_to_dfdict(marginal_2d)

    d = len(marginals_1d)
    upper_inds = get_d_dim_marginal_indices(d, 2)
    assert len(marginals_2d) == len(upper_inds)

    fig, axes = plt.subplots(nrows=d, ncols=d, sharex="col", figsize=figsize)
    _, diag, upper = split_corner_axes(axes)
    lb = 0.125
    tr = 0.9
    fig.subplots_adjust(
        left=lb,
        bottom=lb,
        right=tr,
        top=tr,
        wspace=space_between_axes,
        hspace=space_between_axes,
    )

    for ax in upper:
        ax.axis("off")

    for i, (k, ax) in enumerate(zip(sorted(marginals_1d.keys()), diag)):
        df = marginals_1d[k]
        sns.histplot(
            data=df,
            x=k[0],
            stat="density",
            weights=_set_weight_keyword(df),
            bins=bins,
            ax=ax,
            element="step",
            fill=False,
            color="k",
        )
        if kde:
            sns.kdeplot(
                data=df,
                x=k[0],
                weights=_set_weight_keyword(df),
                ax=ax,
                fill=False,
                color="b",
            )
        ax.tick_params(
            axis="both",
            which="both",
            bottom=False,
            left=False,
            direction="out",
            labelbottom=False,
            labelleft=False,
        )
        if truth is not None:
            ax.axvline(truth[i], color="r")

    # TODO is this sorting always correct to do?
    for k, i in tqdm(zip(sorted(marginals_2d.keys()), upper_inds)):
        a, b = i  # plot array index, upper index (lower index is transpose)
        x, y = k  # marginal index
        ax = axes[b, a]  # targets the lower left corner
        df = marginals_2d[k]

        sns.histplot(
            data=df,
            x=x,
            y=y,
            weights=_set_weight_keyword(df),
            bins=bins,
            ax=ax,
            color="k",
            pthresh=0.01,
        )
        if kde:
            sns.kdeplot(
                data=df,
                x=x,
                y=y,
                weights=_set_weight_keyword(df),
                ax=ax,
                # palette="muted",
                color="b",
                levels=levels,
            )

        if truth is not None:
            ax.axvline(truth[a], color="r")
            ax.axhline(truth[b], color="r")
            ax.scatter(*truth[i, ...], color="r")

        if ylim_lower is None:
            pass
        elif isinstance(ylim_lower[0], (int, float)):
            ax.set_ylim(*ylim_lower)
        elif isinstance(ylim_lower[0], (tuple, list)):
            ax.set_ylim(*ylim_lower[b])
        else:
            raise NotImplementedError(
                "ylim should be a tuple or a list of tuples. Rows are different, columns have the same ylim."
            )

        ax.tick_params(
            axis="both",
            which="both",
            bottom=False,
            left=False,
            # direction="out",
            labelbottom=False,
            labelleft=False,
        )

    # clear all
    for i, axrow in enumerate(axes):
        for j, ax in enumerate(axrow):
            ax.set_xlabel("")
            ax.set_ylabel("")
            if xlim is None:
                pass
            elif isinstance(xlim[0], (int, float)):
                ax.set_xlim(*xlim)
            elif isinstance(xlim[0], (tuple, list)):
                ax.set_xlim(*xlim[j])
            else:
                raise NotImplementedError("xlim should be a tuple or a list of tuples.")

    # bottom row
    for i, ax in enumerate(axes[-1, :]):
        ax.tick_params(
            axis="x",
            which=tickswhich,
            bottom=ticks,
            direction="out",
            labelbottom=ticks and ticklabels,
            labelrotation=labelrotation,
            labelsize=ticklabelsize,
        )
        if labels is not None:
            ax.set_xlabel(labels[i])

    # left column
    for i, ax in enumerate(axes[1:, 0], 1):
        ax.tick_params(
            axis="y",
            which=tickswhich,
            left=ticks,
            direction="out",
            labelleft=ticks and ticklabels,
            labelrotation=labelrotation,
            labelsize=ticklabelsize,
        )
        if labels is not None:
            ax.set_ylabel(labels[i])

    fig.align_labels()
    return fig, axes


if __name__ == "__main__":
    pass



import numpy as np

import swyft

prior_FermiV1 = swyft.Prior(
    {
        "ox": ["uniform", 0.0, 10.0],
        "oy": ["uniform", 0.0, 10.0],
        "a": ["uniform", 1.0, 2.0],
        "p1": ["uniform", 0.0, 0.5],
        "p2": ["uniform", 1.0, 2.0],
    }
)


def simulator_FermiV1(a, ox, oy, p1, p2, sigma=0.1):
    x = np.linspace(-5, 5, 50, 50)
    X, Y = np.meshgrid(x, x)

    diff = np.cos(X + ox) * np.cos(Y + oy) * a + 2

    p = np.random.randn(*X.shape) * p1 - 0.3
    psc = 10 ** p * p2
    n = np.random.randn(*X.shape) * sigma
    mu = diff * 5 + psc + n
    return mu


def model_FermiV1(params):
    mu = simulator_FermiV1(
        params["a"], params["ox"], params["oy"], params["p1"], params["p2"]
    )
    return dict(mu=mu)



import tempfile
from pathlib import Path

import numpy as np
import pytest
import zarr
from dask.distributed import LocalCluster

from swyft.prior import get_uniform_prior
from swyft.store.simulator import DaskSimulator, SimulationStatus, Simulator
from swyft.store.store import Store


def model(params):
    p = np.linspace(-1, 1, 10)
    a, b = params
    mu = p * a + b
    return dict(x=mu)


def broken_model(params):
    raise ValueError("oops!")


def model_multi_out(params):
    p = np.linspace(-1, 1, 10)
    a, b = params
    mu = p * a + b
    mu2 = np.reshape(p * a - b, (2, 5))
    return dict(x1=mu, x2=mu2)


sim = Simulator(model, parameter_names=["a", "b"], sim_shapes=dict(x=(10,)))
sim_multi_out = Simulator(
    model_multi_out, parameter_names=["a", "b"], sim_shapes=dict(x1=(10,), x2=(2, 5))
)
prior = get_uniform_prior(np.zeros(2), np.array([1.0, 0.5]))


class TestStoreIO:
    def test_init_memory_store(self):
        store = Store.memory_store(simulator=sim)
        assert store.v.shape[1] == 2
        assert isinstance(store._zarr_store, zarr.storage.MemoryStore)
        assert isinstance(store._simulator, Simulator)

    def test_init_memory_store_multi_outputs(self):
        store = Store.memory_store(simulator=sim_multi_out)
        assert store.v.shape[1] == 2
        assert {k: v.shape[1:] for k, v in store.sims.items()} == {
            "x1": (10,),
            "x2": (2, 5),
        }

    def test_init_directory_store_multi_outputs(self):
        with tempfile.TemporaryDirectory() as td:
            td_path = Path(td) / "store.zarr"
            store = Store.directory_store(simulator=sim_multi_out, path=td_path)
            assert store.v.shape[1] == 2
            assert {k: v.shape[1:] for k, v in store.sims.items()} == {
                "x1": (10,),
                "x2": (2, 5),
            }
            groups = set(p.name for p in td_path.glob("*") if p.is_dir())
            assert groups == {"samples", "metadata"}
            sims = set(
                p.name for p in (td_path / "samples/sims/").glob("*") if p.is_dir()
            )
            assert sims == {"x1", "x2"}

    def test_init_directory_store_with_existing_path(self):
        with tempfile.TemporaryDirectory() as td:
            with pytest.raises(FileExistsError):
                Store.directory_store(simulator=sim_multi_out, path=td)

    def test_init_directory_store_overwriting_existing_path(self):
        with tempfile.TemporaryDirectory() as td:
            store = Store.directory_store(
                simulator=sim_multi_out, path=td, overwrite=True
            )
            assert store.v.shape[1] == 2
            assert {k: v.shape[1:] for k, v in store.sims.items()} == {
                "x1": (10,),
                "x2": (2, 5),
            }
            td_path = Path(td)
            groups = set(p.name for p in td_path.glob("*") if p.is_dir())
            assert groups == {"samples", "metadata"}
            sims = set(
                p.name for p in (td_path / "samples/sims/").glob("*") if p.is_dir()
            )
            assert sims == {"x1", "x2"}

    def test_memory_store_save(self):
        store = Store.memory_store(simulator=sim_multi_out)
        with tempfile.TemporaryDirectory() as td:
            td_path = Path(td) / "store.zarr"
            store.save(td_path)
            groups = set(p.name for p in td_path.glob("*") if p.is_dir())
            assert groups == {"samples", "metadata"}
            sims = set(
                p.name for p in (td_path / "samples/sims/").glob("*") if p.is_dir()
            )
            assert sims == {"x1", "x2"}

    def test_directory_store_save_to_different_path(self):
        with tempfile.TemporaryDirectory() as td1, tempfile.TemporaryDirectory() as td2:
            td_from = Path(td1) / "store.zarr"
            store = Store.directory_store(simulator=sim_multi_out, path=td_from)
            td_to = Path(td2) / "store.zarr"
            store.save(td_to)
            groups = set(p.name for p in td_to.glob("*") if p.is_dir())
            assert groups == {"samples", "metadata"}
            sims = set(
                p.name for p in (td_to / "samples/sims/").glob("*") if p.is_dir()
            )
            assert sims == {"x1", "x2"}

    def test_directory_store_save_to_same_path_does_not_remove_store(self):
        with tempfile.TemporaryDirectory() as td:
            td_path = Path(td) / "store.zarr"
            store = Store.directory_store(simulator=sim_multi_out, path=td_path)
            store.save(td_path)
            groups = set(p.name for p in td_path.glob("*") if p.is_dir())
            assert groups == {"samples", "metadata"}
            sims = set(
                p.name for p in (td_path / "samples/sims/").glob("*") if p.is_dir()
            )
            assert sims == {"x1", "x2"}

    def test_directory_store_load_existing_store(self):
        with tempfile.TemporaryDirectory() as td:
            td_path = Path(td) / "store.zarr"
            store = Store.directory_store(simulator=sim_multi_out, path=td_path)
            loaded = Store.load(td_path)
            # loaded.set_simulator(sim_multi_out)
            assert np.allclose(loaded.v, store.v)
            assert loaded._zarr_store.path == store._zarr_store.path
            assert loaded.sims.keys() == store.sims.keys()

    def test_directory_store_load_store_from_wrong_paths(self):
        with tempfile.TemporaryDirectory() as td:
            with pytest.raises(KeyError):
                # path exists, keys not found
                Store.load(td)
            with pytest.raises(FileNotFoundError):
                # non-existent path
                td_path = Path(td) / "store.zarr"
                Store.load(td_path)

    def test_directory_store_to_memory(self):
        with tempfile.TemporaryDirectory() as td:
            td_path = Path(td) / "store.zarr"
            store = Store.directory_store(simulator=sim_multi_out, path=td_path)
            loaded = store.to_memory()
            assert isinstance(loaded._zarr_store, zarr.MemoryStore)
            assert np.allclose(loaded.v.shape, store.v.shape)


class TestStoreRun:
    def test_store_add(self):
        store = Store.memory_store(simulator=sim_multi_out)
        store.add(20, prior)
        assert store.sims.x1.shape[0] > 0

    def test_memory_store_simulate(self):
        store = Store.memory_store(simulator=sim_multi_out)
        indices = store.sample(100, prior, add=True)
        ind_sim = indices[:50]
        store.simulate(ind_sim)

        assert store.sims.x1[49].sum() != 0
        assert store.sims.x1[50].sum() == 0

    def test_directory_store_sample(self):
        with tempfile.TemporaryDirectory() as td:
            store = Store.directory_store(
                simulator=sim_multi_out, path=td, overwrite=True
            )
            indices = store.sample(100, prior, add=True)
            assert len(indices) == len(store)

    def test_directory_store_simulate(self):
        with tempfile.TemporaryDirectory() as td:
            store = Store.directory_store(simulator=sim, path=td, overwrite=True)
            ind_sim = store.sample(100, prior, add=True)
            store.simulate(ind_sim)
            assert store.sims.x[:].sum(axis=1).all()

    def test_directory_store_simulate_partial(self):
        with tempfile.TemporaryDirectory() as td:
            store = Store.directory_store(simulator=sim, path=td, overwrite=True)
            ind_sim = store.sample(100, prior, add=True)
            ind_sim = ind_sim[:40]
            store.simulate(ind_sim)
            assert store.sims.x[39].sum() != 0
            assert store.sims.x[40].sum() == 0

    def test_store_lockfile(self):
        with tempfile.TemporaryDirectory() as td:
            store_dir = Store.directory_store(
                simulator=sim, path=td, sync_path=td + ".sync", overwrite=True
            )
            assert store_dir._lock is not None
            assert store_dir._lock.lockfile is None

            store_dir.lock()
            assert store_dir._lock.lockfile is not None

            store_dir.unlock()
            assert store_dir._lock.lockfile is None

    def test_interrupted_simulator_failed(self):
        sim_fail = Simulator(
            broken_model, parameter_names=["a", "b"], sim_shapes={"obs": (2,)}
        )
        store = Store.memory_store(simulator=sim_fail)
        store.add(10, prior)
        store.simulate()
        assert all(store.sim_status[:] == SimulationStatus.FAILED)

    def test_interrupted_dasksimulator_failed(self):
        with tempfile.TemporaryDirectory() as tmpdir:
            with LocalCluster(local_directory=tmpdir) as cluster:
                sim_fail = DaskSimulator(
                    broken_model, parameter_names=["a", "b"], sim_shapes={"obs": (2,)}
                )
                sim_fail.set_dask_cluster(cluster)

                store = Store.memory_store(simulator=sim_fail)
                store.add(10, prior)
                store.simulate()
                assert all(store.sim_status[:] == SimulationStatus.FAILED)



import json
from pathlib import Path

IGNORE = [
    ".ipynb_checkpoints",
    "Video",
]


def create_nblink(notebook_path):
    d = {}
    d["path"] = str(notebook_path)
    d["extra-media"] = []
    return d


def main():
    root = Path(__file__).parent
    # clean directory first
    for old_nblink in root.glob("*.nblink"):
        Path(old_nblink).unlink()

    # add the relevant notebooks and names
    relative_source = Path("../../../notebooks/")
    source = Path(root, relative_source)
    for nbpath in source.glob("*/"):
        if any([ign in str(nbpath) for ign in IGNORE]):
            continue
        else:
            target_path = Path(relative_source, f"{nbpath.stem}.ipynb")
            nblink = create_nblink(target_path)
            linkpath = Path(root, f"{nbpath.stem}.nblink")
            with open(linkpath, "w") as f:
                print("Creating path for", target_path, "Located at", linkpath)
                json.dump(nblink, f)


if __name__ == "__main__":
    main()
